# -*- python -*-
# $RCSfile: colormap.py,v $
# $Revision: 1.20 $
# $Author: langer $
# $Date: 2004/11/12 22:16:30 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# Colormaps map numbers in the range [0,1] to colors

from common import color
from common import registeredclass
from common.IO import parameter
from types import *
from math import *

FloatParameter = parameter.FloatParameter

class ColorMap(registeredclass.RegisteredClass):
    registry = []

ColorMap.tip = "Color maps for <link linkend='RegisteredClass:FilledContourDisplay'><classname>FilledContourDisplay</classname></link>."
ColorMap.discussion = "<para>TODO: Add discussion.</para>"

class ColorMapRegistration(registeredclass.Registration):
    def __init__(self, name, subclass, ordering, params=[], tip=None):
        registeredclass.Registration.__init__(self, name, ColorMap, subclass,
                                              ordering, params=params, tip=tip)
##########################


class GrayMap(ColorMap):
    def __call__(self, x):
        return color.Gray(x)

ColorMapRegistration('Gray', GrayMap, 0, tip="linear gray scale")

class ThermalMap(ColorMap):
    "black->red->yellow->white"
    def __call__(self, x):
        if x < 0.33:
            return color.RGBColor(x/0.33, 0, 0)
        if x < 0.67:
            return color.RGBColor(1, (x-0.33)/0.34, 0)
        return color.RGBColor(1, 1, (x-0.67)/0.33)

ColorMapRegistration('Thermal', ThermalMap, 1,
                     tip="black to red to yellow to white")

import math

class GammaGrayMap(ColorMap):
    def __init__(self, gamma):
        self.gamma = gamma
        self.invgamma = 1.0/gamma
    def __call__(self, x):
        return color.Gray(math.pow(x, self.invgamma))

ColorMapRegistration('GammaGray', GammaGrayMap, 0.1,
                     [FloatParameter('gamma', 1.0,
                                tip="gamma correction")],
                     tip="gray = x^(1/gamma)")

class HSVMap(ColorMap):
    def __init__(self, saturation=1.0, value=1.0, phase_shift=0.0):
        self.saturation = saturation
        self.value = value
        self.phase_shift = phase_shift
    def __call__(self, x):
        return color.HSVColor(hue=x*360+self.phase_shift,
                               saturation=self.saturation,
                               value=self.value)

ColorMapRegistration('HSV', HSVMap, 2,
                     [FloatParameter('saturation', 1.0,
                                     tip='0.0=weak, 1.0=strong'),
                      FloatParameter('value', 1.0,
                                     tip='0.0=dark, 1.0=light'),
                      FloatParameter('phase_shift', 0.0,
                                     tip='initial hue, in degrees')],
                     tip="varying hue, at fixed saturation and value")

# The ROYGBIV map -- although, since it starts from blue, it's really
# the VIBGYOR map.
class SpectralMap(ColorMap):
    def __init__(self, saturation=1.0, value=1.0):
        self.saturation = saturation
        self.value = value
    def __call__(self,x):
        return color.HSVColor(hue=(1.0-x)*240,
                               saturation=self.saturation,
                               value=self.value)

ColorMapRegistration('Spectral', SpectralMap, 2.5,
                     [FloatParameter('saturation', 1.0,
                                     tip='0.0=weak, 1.0=strong'),
                      FloatParameter('value', 1.0,
                                     tip='0.0=dark, 1.0=light')],
                     tip="blue, green, yellow, orange, red")


class TequilaSunrise(ColorMap):
    def __call__(self,x):
        return color.RGBColor(1.0, 0.86667*sqrt(x), 0.0)

ColorMapRegistration('TequilaSunrise', TequilaSunrise, 3.,
       tip="It's another Tequila sunrise/Stirrin' slowly 'cross the sky...")


# For parameters containing these, use the appropriate RegisteredParameter.
