// -*- C++ -*-
// $RCSfile: pixelsetboundary.C,v $
// $Revision: 1.7 $
// $Author: reida $
// $Date: 2005/02/28 22:13:07 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modifed
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>

#include "common/pixelsetboundary.h"
#include "common/cmicrostructure.h"
#include <map>


bool PixelBdySegment::operator<(const PixelBdySegment &other) const {
  if (!other.horizontal && horizontal) return true;
  if (!horizontal && other.horizontal) return false;
  
  if (other.pixel < pixel) return true;
  return false;
}


// This both assigns coordinates and directs the segment, but it
// still lives in integer-land.
ICoord PixelBdySegment::coordinate() const {
  ICoord lowerleft = pixel;

  if (horizontal) {
    ICoord lowerright = lowerleft + ICoord(1, 0);
    if (local) {
      start = lowerleft;
      end = lowerright;
    }
    else { 
      end = lowerleft; 
      start = lowerright;
    }
  }
  else {
    ICoord upperleft = lowerleft + ICoord( 0, 1 );
    if (local) {
      start = upperleft;
      end = lowerleft;
    }
    else {
      start = lowerleft;
      end = upperleft;
    }
  }
  coordinated = true;
  return start;
}

void PixelBdySegment::extend() {
  if (horizontal) {
    if (local) 
      end += ICoord(1,0);
    else
      end -= ICoord(1,0);
  }
  else {
    if (local) 
      end -= ICoord(0,1);
    else
      end += ICoord(0,1);
  }
}


// Set the boundary points in physical space.
void PixelBdySegment::set_floats(const Coord &delta) const {
  if (horizontal) {
    fp_level = start(1)*delta(1);  // Segment has constant y
    fp_start = start(0)*delta(0);
    fp_end = end(0)*delta(0);
  }
  else {
    fp_level = start(0)*delta(0);  // Segment has constant x.
    fp_start = start(1)*delta(1);
    fp_end = end(1)*delta(1);
  }

  if (fp_start < fp_end) {
    fp_first = fp_start;
    fp_last = fp_end;
  }
  else {
    fp_first = fp_end;
    fp_last = fp_start;
  }

  floated = true;
}
  

// Find intersection -- if "true" is returned, then the passed-in
// PixelBdyIntersection has been filled in with intersection data.
// Otherwise, it has not been modified.
bool PixelBdySegment::find_intersection(const Coord &p1, 
					const Coord &p2,
					PixelBdyIntersection &pbi) const {
  if (horizontal) {
    if ( ( p1(0) < fp_first ) && (p2(0) < fp_first ) )
      return false;
    if ( ( p1(0) > fp_last ) && (p2(0) > fp_last ) )
      return false;
    if ( ( p1(1) < fp_level) && ( p2(1) < fp_level) )
      return false;
    if ( ( p1(1) > fp_level) && ( p2(1) > fp_level) )
      return false;
    // If we made it this far, then an intersection is possible.

    // Co-linear case is covered by the "grazing" cases -- ignore here.
    if (p1(1)==p2(1)) 
      return false; 

    // If end-point of (p1,p2) grazes us, that might be an
    // intersection.  The logic bomb after the && is an interiority
    // check on p2(0) -- it has to be between the endpoints, but we
    // don't know what order the endpoints come in.
    if ( (p2(1)==fp_level) && ((p2(0)>=fp_first) && (p2(0)<=fp_last)) ) {
      // If we are local, then we go left-to-right, and the interior
      // of the pixel region is "above" us.  If the other endpoint of
      // the passed-in segment is also above us, then we are entering
      // the element here.
      if ((local && (p1(1) > fp_level)) || (!local && (p1(1) < fp_level))) {
	pbi.location(0) = p2(0);
	pbi.location(1) = fp_level;
	pbi.entry = true;
	return true;
      }
      // If neither case is true, then we don't know -- define this as
      // a non-intersection.
      else {
	return false;
      }
    } 
      
    
    // If the start-point of (p1,p2) grazes us, that might be an intersection.
    if ( (p1(1)==fp_level) && ((p1(0)>=fp_first) && (p1(0)<=fp_last)) ) {
      // If we are local then we go left-to-right, and the interior of
      // the pixel region is "above" us -- if the other endpoint of
      // the passed-in segment is also above us, then this is where
      // the pixel boundary exits the element.
      if ((local && (p2(1) > fp_level )) || (!local && (p2(1) < fp_level))) {
	pbi.location(0) = p1(0);
	pbi.location(1) = fp_level;
	pbi.entry = false; // "outgoing" intersection at p1.
	return true;
      }
      // If neither case is true, then it's not an intersection.
      else {
	return false; 
      }
    }

    // Nontrivial crossing -- find it. 
    double alpha = (fp_level - p1(1)) / ( p2(1)-p1(1) );
    double x = p1(0) + alpha*(p2(0)-p1(0));
    
    // If we are horizontal and local, then we go left-to-right.  If
    // the end-point of the element boundary segment is above us, then
    // the element interior is on the left, and we are exiting it.
    if ( x >= fp_first && x <= fp_last ) {
      pbi.location(0) = x;
      pbi.location(1) = fp_level;
      if ((local && (p2(1) > fp_level )) || (!local && (p2(1) < fp_level))) {
	pbi.entry = false; // "incoming" at (x, fp_level)
	return true;
      } 
      else {
	pbi.entry = true; // "outgoing" at (x, fp_level)
	return true;
      }
    }
    
    return false;  // No intersection found.
  }
  else {  // Vertical.
    // Basic bounds check
    if ( ( p1(1) < fp_first ) && (p2(1) < fp_first ) )
      return false;
    if ( ( p1(1) > fp_last ) && (p2(1) > fp_last ) )
      return false;
    if ( ( p1(0) < fp_level) && ( p2(0) < fp_level) )
      return false;
    if ( ( p1(0) > fp_level) && ( p2(0) > fp_level) )
      return false;
    // If we made it this far, then an intersection is possible.

    if (p1(0)==p2(0)) // Co-linear case covered by "grazing" cases.
      return false; 


    // If end-point of (p1, p2) grazes us, that might be an intersection.
    if ( (p2(0)==fp_level) && ((p2(1)>=fp_first) && (p2(1)<=fp_last)) ) {
      // If we are local, then we go top-to-bottom, and the interior
      // of the pixel group is on the right.  If the start point of
      // (p1,p2) is also on the right, then the element interior is
      // below, and we are entering it.
      if ((local && (p1(0) > fp_level)) || (!local && (p1(0) < fp_level))) {
	pbi.location(0) = fp_level;
	pbi.location(1) = p2(1);
	pbi.entry = true;
	return true;
      }
      else {
	return false;
      }
    }

    
    // If the start-point of (p1,p2) grazes us, possible intersection.
    if ((p1(0)==fp_level) && ((p1(1)>=fp_first) && (p1(1)<=fp_last)) ) {
      // If we are local, then we go top-to-bottom, and the interior
      // of the pixel group is on the right.  If the end point of (p1,
      // p2) is also on the right, then the element interior is above,
      // and we are exiting it.
      if ((local && p2(0) > fp_level) || (!local && (p2(0) < fp_level))) {
	pbi.location(0) = fp_level;
	pbi.location(1) = p1(1);
	pbi.entry = false;
	return true;
      }
      else {
	return false;
      }
    }

    // Nontrivial crossing -- find it. 
    double alpha = (fp_level - p1(0)) / ( p2(0)-p1(0) );
    double y = p1(1) + alpha*(p2(1)-p1(1));
    
    // If we are vertical and local, then we go top-to-bottom.  If the
    // end-point of the element segment is to the right, then the
    // element interior is above us, and we are exiting it.
    if  (y >= fp_first && y <= fp_last)  {
      pbi.location(0) = fp_level;
      pbi.location(1) = y;
      if ((local && (p2(0) > fp_level )) || (!local && (p2(0) < fp_level))) {
	pbi.entry = false; 
	return true;
      } 
      else {
	pbi.entry = true;
	return true;
      }
    }
    return false;  // No intersection found.
  }
}


Coord PixelBdySegment::start_pt() const {
  if (horizontal)
    return Coord(fp_start, fp_level);
  return Coord(fp_level, fp_start);
}

Coord PixelBdySegment::end_pt() const {
  if (horizontal) 
    return Coord(fp_end, fp_level);
  return Coord(fp_level, fp_end);
}


std::ostream& operator<<(std::ostream &o, const PixelBdySegment &pbs) {
  if (pbs.floated) {
    if (pbs.horizontal) {
      o << "(" << pbs.fp_start << ", " << pbs.fp_level << 
	") -> (" << pbs.fp_end << ", " << pbs.fp_level << ")";
    }
    else {
      o << "( " << pbs.fp_level << ", " << pbs.fp_start << 
	") -> (" << pbs.fp_level << ", " << pbs.fp_end << ")";
    }
  }
  else if (pbs.coordinated) {
    o << pbs.start << ", " << pbs.end;
  }
  else {
    o << (pbs.horizontal ? " Horizontal " : " Vertical ") << 
      (pbs.local ? " Local " : " Nonlocal ") << pbs.pixel;
  }
  return o;
}



// Create the boundary segments for this pixel, and add them to the
// map of all pixel segments.  PixelBdySegment constructor args are
// pixel ICoord, horizontality, and locality.
void PixelSetBoundary::add_pixel(const ICoord &px) {
  segments[PixelBdySegment(px, true, true)] += 1;
  segments[PixelBdySegment(px, false, true)] += 1;

  ICoord neighbor = px + ICoord(0,1);
  segments[PixelBdySegment(neighbor, true, false)] += 1;
  
  neighbor = px + ICoord(1,0);
  segments[PixelBdySegment(neighbor, false, false)] += 1;
}



PixelSetBoundary::PixelSetBoundary(const CMicrostructure* ms) :
  segments(), microstructure(ms), loopset(0) {
  ms_delta = microstructure->sizeOfPixels();
}

typedef std::pair<ICoord, PixelBdySegment> SegByCoord;


void PixelSetBoundary::find_boundary() {

  // Coord-keyed map of segments.
  CoordMap coordmap;

  // Remove the segments which occur twice -- they are interior.
  SegMap::iterator i = segments.begin();
  while(i!=segments.end()) {
    if ((*i).second == 2) {
      SegMap::iterator j = i;
      ++i;
      segments.erase(j);
    }
    else {
      ++i;
    }
  }

  // At this point, the keys in the "segments" map all have value "1",
  // and are grouped, but not in a useful way.  The geometry of the
  // pixel array guarantees that we have at least four segments.
  // Put them in a new multimap, indexed by their start coords.
  SegMap::iterator s = segments.begin();
  // Avoid making a copy of the PixelBdySegment.
  while(s!=segments.end()) {
    const PixelBdySegment &pbs = (*s).first;
    ICoord segment_start = pbs.coordinate();
    coordmap.insert( SegByCoord(segment_start, pbs ) );
    SegMap::iterator t = s;
    ++s;
    segments.erase(t);
  } 

  // Now we have a Coord-indexed multimap of segment objects, and
  // furthermore, all of the indices are the start of a coordinated
  // segment.  Now, construct the loops, merrily erasing as we go.
  
  while (coordmap.size() != 0) {
    loopset.push_back(find_loop(coordmap));
  }

  // Optimization: Clean up redundant segments from the loops.
  for(LoopVec::iterator ell = loopset.begin(); ell!=loopset.end(); ++ell) {
    clean_loop(*ell);
  }

  // Final pass -- set-floats in all the segments.
  for(LoopVec::iterator ell = loopset.begin(); ell!=loopset.end(); ++ell) {
    for(std::vector<PixelBdySegment>::iterator x = (*ell).begin(); 
	x!=(*ell).end(); ++x) {
      (*x).set_floats(ms_delta);
    }
  }
}


typedef std::pair<CoordMap::iterator, CoordMap::iterator> CoordMapRange;
 
// Finds a loop in the passed-in coordinate map, and returns it.
// Removes the relevant segments from the cm, also, which is why it's
// a reference and not const.
std::vector<PixelBdySegment> PixelSetBoundary::find_loop(CoordMap &cm) {

  std::vector<PixelBdySegment> result;
  CoordMap::iterator current = cm.begin();
  
  bool done = false;
  while (!done) {
    PixelBdySegment current_seg = (*current).second;
    result.push_back(current_seg);
    ICoord next = current_seg.end;
    cm.erase(current);

    int n = cm.count(next);
    
    if (n==0)
      done=true;
    else if (n==1) {
      current = cm.lower_bound(next);
    }
    else {
      CoordMapRange range = cm.equal_range(next);
      for (CoordMap::iterator i = range.first; i!=range.second; ++i) {
	current = i;
	if (current_seg.horizontal && !(*i).second.horizontal &&
	    (current_seg.local != (*i).second.local) )
	  break;
	if (!current_seg.horizontal && (*i).second.horizontal &&
	    (current_seg.local == (*i).second.local) )
	  break;
      }
    }
  }
  return result;
}


// Cleans up the loop in-place, by extending segments for which the
// adjacent segment is just a continuation of the current segment.
void PixelSetBoundary::clean_loop(std::vector<PixelBdySegment>& loop) {

  std::vector<PixelBdySegment>::iterator current = loop.begin();
  std::vector<PixelBdySegment>::iterator i = current;
  i++;
  // STL semantics are such that iterators preceding the erasure point
  // remain valid after an erasure, but those afterwards do not, so
  // current < i is important.
  while(i!=loop.end()) {
    if ((*i).horizontal == (*current).horizontal) {
      (*current).extend();
      loop.erase(i);
      i = current;
      i++;
    }
    else {
      current = i;
      i++;
    }
  }
}

							 
							 
//\\//\\//\\//\\//\\//\\//\\//\\//\\//\\//\\//\\//\\//\\//\\//\\//\\

typedef std::vector<PixelBdySegment> Loop;
typedef std::vector<Loop> LoopSet;


std::ostream& operator<<(std::ostream &o, const PixelSetBoundary &psb) {
  for(LoopSet::const_iterator i = psb.loopset.begin(); 
      i!=psb.loopset.end(); ++i) {
    o << "Loop: " << std::endl;
    for(Loop::const_iterator j = (*i).begin(); j!=(*i).end(); ++j) {
      o << "Segment : " << (*j) << std::endl;
    }
  }
  return o;
}
