# -*- python -*-
# $RCSfile: boundarycondition.py,v $
# $Revision: 1.58 $
# $Author: langer $
# $Date: 2005/03/02 20:29:51 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

from SWIG.common import switchboard
from common import debug
from common import labeltree
from common import utils
from common.IO import automatic
from common.IO import oofmenu
from common.IO import parameter
from common.IO import reporter
from common.IO import whoville
from common.IO import xmlmenudump
from engine import bdycondition
from engine import skeletoncontext
from engine.IO import meshmenu
from engine.IO import meshparameters
import engine.mesh
import engine.profile
import types

bcmenu = oofmenu.OOFMenuItem(
    'Boundary_Conditions',
    cli_only=1,
    discussion=xmlmenudump.loadFile(
    'DISCUSSIONS/engine/IO/menu-bc.xml'))

meshmenu.meshmenu.addItem(bcmenu)


def _buildbc(menuitem, name, mesh, condition):
    condition.add_to_mesh(name, mesh)
    switchboard.notify("mesh changed", engine.mesh.meshes[mesh])

def bcnameResolver(param, name):
    if param.automatic():
        basename = 'bc'
    else:
        basename = name
    meshname = param.group['mesh'].value
    if meshname is not None:
        meshpath = labeltree.makePath(meshname)
        meshcontext = engine.mesh.meshes[meshpath]
        return meshcontext.uniqueBCName(basename)

bcmenu.addItem(oofmenu.OOFMenuItem('New',
                                   callback=_buildbc,
                                   params=parameter.ParameterGroup(
    parameter.AutomaticNameParameter('name',
                                     value=automatic.automatic,
                                     tip="Name of the new boundary condition",
                                     resolver=bcnameResolver),
    whoville.WhoParameter('mesh', engine.mesh.meshes,
                          tip='Mesh to which to apply the boundary condition.'),
    parameter.RegisteredParameter("condition", bdycondition.BC,
                                  tip="The new boundary condition.")),
                                   help='Create a new boundary condition.'
    ))


def _bcrename(self, mesh, bc, name):
    meshcontext = engine.mesh.meshes[mesh]
    meshcontext.renameBdyCondition(bc, name)

bcmenu.addItem(oofmenu.OOFMenuItem(
    "Rename",
    callback=_bcrename,
    params=parameter.ParameterGroup(
    whoville.WhoParameter("mesh", engine.mesh.meshes,
                          tip=parameter.emptyTipString),
    parameter.StringParameter('bc', tip='Old name of the boundary condition'),
    parameter.StringParameter('name', tip='New name of the boundary condition')
##     parameter.AutomaticNameParameter('name', value=automatic.automatic,
##                                      tip="New name for the bc",
##                                      resolver=bdycondition.bcnameResolver)
    ),
    help="Give a boundary condition a new name."
    ))

# Remove a boundary condition from the given mesh.
def _bcremove(self, mesh, name):
    meshcontext = engine.mesh.meshes[mesh]
    meshcontext.rmBdyCondition(name)
    switchboard.notify("mesh changed", engine.mesh.meshes[mesh])
    
bcmenu.addItem(oofmenu.OOFMenuItem(
    "Delete",
    callback=_bcremove,
    params = [whoville.WhoParameter('mesh', engine.mesh.meshes,
                                    tip="Name of the mesh."),
              parameter.StringParameter('name', tip="Name of the boundary condition to be deleted.")],
    help="Remove this condition from the mesh."
    ))


# Copy a boundary condition to a new boundary (for possibly another mesh).
def _bccopy(self, current, mesh, bc, name, boundary):
    meshcontext = engine.mesh.meshes[current]
    bndycond = meshcontext.getBdyCondition(bc)
    newbc = bndycond.copy(boundary)
    newbc.add_to_mesh(name, mesh)
    switchboard.notify("mesh changed", engine.mesh.meshes[mesh])
    
bcmenu.addItem(oofmenu.OOFMenuItem(
    "Copy",
    callback=_bccopy,
    help="Copy a boundary condition from one mesh to another.",
    params=parameter.ParameterGroup(
    whoville.WhoParameter("current", engine.mesh.meshes,
                          tip="Copy the boundary condition from this mesh."),
    whoville.WhoParameter("mesh", engine.mesh.meshes,
                          tip="Copy the boundary condition to this mesh."),
    parameter.StringParameter("bc", tip="Boundary condition being copied"),
    parameter.AutomaticNameParameter("name", value=automatic.automatic,
                                     resolver=bcnameResolver,
                                     tip="Name of the new boundary condition."),
    meshparameters.MeshBoundaryParameter(
    "boundary",tip="Boundary to which the copied BC will apply."))))

# Copy all the boundary conditions from one mesh to another, copying
# them to the same boundaries, of course.
def _bccopyall(self, current, mesh):
    meshcontext = engine.mesh.meshes[current]
    allbcs = meshcontext.allBoundaryConds()
    for (name, cond) in allbcs:
        newbc = cond.copy(cond.boundary)
        newbc.add_to_mesh(name, mesh)
    switchboard.notify("mesh changed", engine.mesh.meshes[mesh])
    
bcmenu.addItem(oofmenu.OOFMenuItem(
    "Copy_All",
    callback=_bccopyall,
    params=[
    whoville.WhoParameter("current", engine.mesh.meshes,
                          tip="Copy all the boundary conditions to a mesh."),
    whoville.WhoParameter("mesh", engine.mesh.meshes,
                          tip="Target mesh for the boundary conditions.")],
    help="Copy all the boundary conditions from one mesh to another."))


def _bcedit(menuitem, name, mesh, condition):
    meshctxt = engine.mesh.meshes[mesh]
    meshctxt.rmBdyCondition(name)
    condition.add_to_mesh(name, mesh)
    switchboard.notify("mesh changed", engine.mesh.meshes[mesh])    
    
bcmenu.addItem(oofmenu.OOFMenuItem(
    "Edit",
    callback=_bcedit,
    params = [parameter.StringParameter("name",
                                        tip="Name for the boundary condition."),
              # No resolver, name must exist
              whoville.WhoParameter("mesh", engine.mesh.meshes,
                                    tip=parameter.emptyTipString),
              parameter.RegisteredParameter("condition", bdycondition.BC,
                                            tip=parameter.emptyTipString)],
    help="Change the attributes of a boundary condition."))

                 
    


##############################
### Similar for Profiles.
##############################

pfmenu = oofmenu.OOFMenuItem('Profiles', cli_only=1,
                             discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu-profile.xml'))

meshmenu.meshmenu.addItem(pfmenu)

def pfnameResolver(param, name):
    if param.automatic():
        basename = 'profile'
    else:
        basename = name
    return engine.profile.AllProfiles.uniqueName(basename)

def newNamedProfile(menuitem, name, profile):
    profile.set_name(name)

pfmenu.addItem(oofmenu.OOFMenuItem(
    'New',
    callback=newNamedProfile,
    params=[
    parameter.AutomaticNameParameter('name',
                                     value=automatic.automatic,
                                     tip="Name for the new Profile.",
                                     resolver=pfnameResolver),
    parameter.RegisteredParameter('profile', engine.profile.CoreProfile,
                                  tip='The new Profile.')
    ],
    help='Construct a new named Profile.',
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu-newprofile.xml')
                                   ))


# Remove a profile from the profile manager.
def _pfremove(self, name):
    pf = engine.profile.AllProfiles[name]
    if pf.deletable():
        del engine.profile.AllProfiles[name]
    else:
        reporter.warn("Profile '%s' cannot be deleted because it is still being used." % name)
    
pfmenu.addItem(oofmenu.OOFMenuItem(
    "Delete",
    help="Delete a profile from the system.",
    callback=_pfremove,
    params = [parameter.StringParameter('name', tip="Name of the profile to be deleted.")],
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu-deleteprofile.xml')
    ))


def _pfrename(self, profile, name):
    engine.profile.AllProfiles.rename(profile, name)

pfmenu.addItem(oofmenu.OOFMenuItem(
    "Rename",
    help="Rename a profile.",
    callback=_pfrename,
    params = [
    parameter.StringParameter('profile', tip="Old name of the profile."),
    parameter.StringParameter('name', tip='New name of the profile')],
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu-renameprofile.xml')
    ))


def _pfcopy(self, profile, name):
    old_profile = engine.profile.AllProfiles[profile]
    new_profile = old_profile.clone()
    new_profile.set_name(name)

pfmenu.addItem(oofmenu.OOFMenuItem(
    "Copy",
    help="Make a copy of a profile.",
    callback=_pfcopy,
    params = [
    parameter.StringParameter('profile', tip="Name of the profile to be copied."),
    parameter.AutomaticNameParameter('name', resolver=pfnameResolver,
                                     tip="Name for the new profile.") ],
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu-copyprofile.xml')
    ))


def _pfedit(self, name, profile):
    engine.profile.AllProfiles.replaceProfile(name, profile)
    
pfmenu.addItem(oofmenu.OOFMenuItem(
    "Edit",
    help="Replace a profile with a new profile.",
    callback=_pfedit,
    params = [
    parameter.StringParameter("name", tip="Name of the profile to be replaced."),
    # No resolver, name must exist
    parameter.RegisteredParameter("profile", engine.profile.CoreProfile,
                                  tip="The replacement profile.")],
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu-editprofile.xml')
    ))

