# -*- python -*-
# $RCSfile: isocijkl.py,v $
# $Revision: 1.22 $
# $Author: langer $
# $Date: 2004/11/15 15:35:39 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# The IsoCijkl registered-class hierarchy.  IsoCijkl hierarchy members
# are convertible registered class objects, which provide access to the
# various different ways of parametrizing an isotropic fourth-rank
# index-exchange-symmetric tensor in three dimensions.

from common import registeredclass
from common.IO import parameter
from common.IO import xmlmenudump
FloatParameter = parameter.FloatParameter

##from common import utils
import types


# Base class.  Should be able to answer questions about its values
# in different representations, using the member *instance*'s to_base
# functions.  These should use the code from the registration's
# to_base and from_base routines, but operate on the attributes of the
# Cijkl instance, and not the default parameters.  These are used
# by the properties internally to set their Cijkl objects for
# computation.
class IsotropicRank4Tensor(registeredclass.ConvertibleRegisteredClass):
    registry = []

# Base value type for the IsotropicRank4Tensor.  This is necessary for
# convertible types, because the from_base conversion routine needs to
# have some confidence that it's getting a value it can cope with,
# which it does by type-checking against this type.
class IsoCijklValueBase:
    def __init__(self,c11,c12):
        self.c11 = c11
        self.c12 = c12
    def __repr__(self):
        return "IsoCijklValueBase(c11=%f, c12=%f)" % (self.c11, self.c12)


# Registration object hosts the "call points" for to_base
# routines.  These are used by the GUI when switching types.
class IsoCijklRegistration(registeredclass.ConvertibleRegistration):
    def __init__(self, name, subclass, ordering,
                 from_base, to_base, params=[], tip=None):
        registeredclass.Registration.__init__(self, name, IsotropicRank4Tensor,
                                              subclass,
                                              ordering, from_base=from_base,
                                              to_base=to_base, params=params,
                                              tip=tip)
    def getParamValuesAsBase(self):
        return self.to_base(self)
    def __repr__(self):
        return "IsoCijklRegistration('%s', %s, ordering=%s, params=%s)" % \
               ( self.name, self.subclass.__name__, `self.ordering`,
                 self.params)
    

#########################################################################
#########################################################################

# Conversion for the symmetric matrix representation of the Cijkl
# matrix.  Same internal data representation as the base value type.

class IsotropicRank4TensorCij(IsotropicRank4Tensor):
    def __init__(self, c11, c12):
        # Set the parameters in the registration object.  This is
        # redundant (but harmless) when the object is instanced from
        # the registration.
        self.c11 = c11
        self.c12 = c12

# vlist, if provided, should be an indexable with vlist[0]=c11,
# vlist[1]=c12.  
def _cij_to_base(cijkl_reg, vlist = None):
    vset = vlist or cijkl_reg.getParamValues() # Gets c11 and c12.
    return IsoCijklValueBase(c11=vset[0], c12=vset[1])

def _cij_from_base(cijkl_reg, base):
    if isinstance(base, IsoCijklValueBase):
        cijkl_reg.setDefaultParams([base.c11,base.c12])

IsoCijklRegistration('Cij', IsotropicRank4TensorCij, 100,
                     from_base=_cij_from_base,
                     to_base=_cij_to_base,
                     params=[FloatParameter('c11', 10.0),
                             FloatParameter('c12', 5.0)] )


##utils.OOFdefine('CijIsoCijkl',CijIsoCijkl)

###########################################################################
###########################################################################

# Lame coefficients for isotropic elasticity.
class IsotropicRank4TensorLame(IsotropicRank4Tensor):
    def __init__(self, lmbda, mu):
        self.lmbda = lmbda
        self.mu = mu

# vlist should again be an indexable with vlist[0]=lmbda, vlist[1]=mu
def _lame_to_base(lame_reg, vlist=None):
    (lmbda, mu) = vlist or lame_reg.getParamValues()
    c11 = lmbda + 2.0*mu
    c12 = lmbda
    return IsoCijklValueBase(c11=c11, c12=c12)

def _lame_from_base(lame_reg, base):
    if isinstance(base, IsoCijklValueBase):
        lmbda = base.c12
        mu = (base.c11-base.c12)/2.0
        lame_reg.setDefaultParams([lmbda,mu])

IsoCijklRegistration('Lame', IsotropicRank4TensorLame, 101,
                     from_base=_lame_from_base,
                     to_base=_lame_to_base,
                     params=[FloatParameter('lmbda', 10.0),
                             FloatParameter('mu', 5.0)] )


###########################################################################
###########################################################################

# Young modulus and Poisson ratio.
class IsotropicRank4TensorEnu(IsotropicRank4Tensor):
    def __init__(self, young, poisson):
        self.young = young
        self.poisson = poisson

# vlist, if provided, is indexable.
def _enu_to_base(enu_reg, vlist=None):
    (young, poisson) = vlist or enu_reg.getParamValues()
    c11 = young*(1.0-poisson)/((1.0+poisson)*(1.0-2.0*poisson))
    c12 = young*poisson/((1.0+poisson)*(1.0-2.0*poisson))
    return IsoCijklValueBase(c11=c11, c12=c12)

def _enu_from_base(enu_reg, base):
    if isinstance(base, IsoCijklValueBase):
        c11 = base.c11
        c12 = base.c12
        young = ((c11-c12)*(c11+2.0*c12)/(c11+c12))
        poisson = c12/(c11+c12)
        enu_reg.setDefaultParams([young, poisson])

IsoCijklRegistration('E and nu', IsotropicRank4TensorEnu,
                     102, from_base=_enu_from_base,
                     to_base=_enu_to_base,
                     params=[FloatParameter('young', 31.6),
                             FloatParameter('poisson', 14.7)] )

##########################################################################
##########################################################################

# Bulk and shear modulus scheme.
class IsotropicRank4TensorBulkShear(IsotropicRank4Tensor):
    def __init__(self, bulk, shear):
        self.bulk = bulk
        self.shear = shear

def _bs_to_base(bs_reg, vlist=None):
    (bulk, shear) = vlist or bs_reg.getParamValues()
    c11 = bulk+(4.0/3.0)*shear
    c12 = bulk-(2.0/3.0)*shear
    return IsoCijklValueBase(c11=c11, c12=c12)

def _bs_from_base(bs_reg, base):
    if isinstance(base, IsoCijklValueBase):
        bulk = (1.0/3.0)*(base.c11+2.0*base.c12)
        shear = (base.c11-base.c12)/2.0
        bs_reg.setDefaultParams([bulk, shear])

IsoCijklRegistration('Bulk and Shear', IsotropicRank4TensorBulkShear, 103,
                     from_base=_bs_from_base,
                     to_base=_bs_to_base,
                     params=[FloatParameter('bulk', 31.6),
                             FloatParameter('shear', 14.7)] )

###########################################################################
###########################################################################

# Define a special Parameter type so that it can have special widgets
# assigned to it in GUI mode.

class IsotropicRank4TensorParameter(parameter.ConvertibleRegisteredParameter):
    def __init__(self, name, value=None, default=None, tip=None):
        parameter.ConvertibleRegisteredParameter.__init__(
            self, name, IsotropicRank4Tensor, value=value, default=default,
            tip=tip)
    def clone(self):
        return IsotropicRank4TensorParameter(self.name, self.value,
                                             self.default, self.tip)

######################

## XML Documentation

IsotropicRank4Tensor.tip = "Representations of an isotropic 4th rank tensor."

IsotropicRank4Tensor.discussion = xmlmenudump.loadFile(
    'DISCUSSIONS/engine/IO/isocijkl.xml')
