# -*- python -*-
# $RCSfile: skeleton.py,v $
# $Revision: 1.195 $
# $Author: langer $
# $Date: 2005/03/01 19:12:49 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov.

# A Skeleton contains the geometry of a finite element mesh, without
# any of the details.  It has lists of SkeletonElements,
# SkeletonNodes, and SkeletonEdgeBoundary (made up of SkeletonEdges).
# but no intermediate nodes, materials, or shape functions.


from SWIG.common import ooferror
from SWIG.common import coord
from SWIG.common import timestamp
from SWIG.engine import femesh
from SWIG.common import switchboard
from SWIG.common.IO import progressbar
from common import array2d
from common import debug
from common import enum
from common import primitives
from common import registeredclass
from common import utils
from common import microstructure
from common.IO import parameter
from common.IO import reporter
from engine import materialmanager
from engine import skeletonboundary
from engine import skeletoncontext
from engine import skeletondiff
from engine import skeletonelement
from engine import skeletongroups
from engine import skeletonnode
from engine import skeletonsegment
from engine import skeletonselectable
import math
import random
import types
import weakref

Parameter = parameter.Parameter
Registration = registeredclass.Registration

SkeletonNode = skeletonnode.SkeletonNode
SkeletonEdgeBoundary = skeletonboundary.SkeletonEdgeBoundary
ExteriorSkeletonEdgeBoundary = skeletonboundary.ExteriorSkeletonEdgeBoundary
SkeletonPointBoundary = skeletonboundary.SkeletonPointBoundary
ExteriorSkeletonPointBoundary = skeletonboundary.ExteriorSkeletonPointBoundary
SkeletonEdge = skeletonsegment.SkeletonEdge
SkeletonQuad = skeletonelement.SkeletonQuad
SkeletonTriangle = skeletonelement.SkeletonTriangle
SkeletonSegment = skeletonsegment.SkeletonSegment

switchboard = switchboard.switchboard

# Triangular skeleton arrangements

class Arrangement(
    enum.EnumClass(('conservative', 'leaning to the right'),
                   ('liberal', 'leaning to the left'),
                   ('moderate', 'going both ways'),
                   ('middling', 'going both ways, the other way'),
                   ('anarchic', 'going every which way'))):
    pass
utils.OOFdefine('Arrangement', Arrangement)

# Known arrangements.
conservative = Arrangement('conservative')
liberal = Arrangement('liberal')
moderate = Arrangement('moderate')
anarchic = Arrangement('anarchic')
middling = Arrangement('middling')

Arrangement.tip = "Arrangement for triangular initial Skeleton."
Arrangement.discussion = "<para>TODO: Add discussion.</para>"

class SkeletonGeometry(registeredclass.RegisteredClass):
    registry = []
    def __init__(self, type):
        self.type = type

SkeletonGeometry.tip = "Element shape of initial Skeleton."
SkeletonGeometry.discussion = "<para>TODO: Add discussion.</para>"

class QuadSkeleton(SkeletonGeometry):
    def __init__(self):
        SkeletonGeometry.__init__(self, 'quad')
    def __call__(self, m, n, microStructure, preset_homog=0):
        pBar = progressbar.getProgress()
        """
        Create a skeleton of quadrilateral elements, n rows by m columns.  The
        width and height of the entire skeleton are h and w.
        """
        skel = Skeleton(microStructure)
        btm = skel.getEdgeBoundary('bottom', exterior=1)
        top = skel.getEdgeBoundary('top', exterior=1)
        rgt = skel.getEdgeBoundary('right', exterior=1)
        lft = skel.getEdgeBoundary('left', exterior=1)

        btmlft = skel.getPointBoundary('bottomleft', exterior=1)
        btmrgt = skel.getPointBoundary('bottomright', exterior=1)
        toplft = skel.getPointBoundary('topleft', exterior=1)
        toprgt = skel.getPointBoundary('topright', exterior=1)

        ## create nodes and selected point boundaries.
        dx = (skel.MS.size()[0]*1.0)/m   # Promote numerators to floating-point.
        dy = (skel.MS.size()[1]*1.0)/n
        pBar.set_message("allocating nodes")
        tot_items = (m + 1)*(n + 1)
        for i in range(n+1):
            for j in range(m+1):
                node = skel.newNode(j*dx, i*dy)
                if i==0 and j==0:
                    btmlft.addNode(node)
                if i==0 and j==m:
                    btmrgt.addNode(node)
                if i==m and j==0:
                    toplft.addNode(node)
                if i==m and j==m:
                    toprgt.addNode(node)
                if pBar.query_stop():
                    pBar.set_failure()
                    pBar.set_message("Failed")
                    return None
                else:
                    pBar.set_progress(1.0*(i*(n+1)+j +1)/tot_items)

        ## create elements and edges
        pBar.set_message("creating elements from nodes")
        tot_items = m*n
        for i in range(n):                  # loop over rows of elements
            for j in range(m):              # loop over columns of elements
                ll = i*(m+1) + j            # lower left node index
                lr = i*(m+1) + j+1          # lower right
                ur = (i+1)*(m+1) + j+1      # upper right
                ul = (i+1)*(m+1) + j        # upper left
                el = skel.newElement([skel.nodes[ll],skel.nodes[lr],
                                      skel.nodes[ur],skel.nodes[ul]])
                # Simple skeleton -- set the homogeneity to "1".
                if preset_homog:
                    el.cachedHomogeneity.value = 1.0
                    el.cachedDominantPixel = microStructure.pixelFromPoint(
                        el.repr_position())
                
                # Element constructors make the segments, we can use them
                # to make the edges and add them to the boundaries.
                if j == 0:                 
                    segment = skel.segments[
                        utils.canonicalorder(skel.nodes[ul], skel.nodes[ll])]
                    edge = SkeletonEdge(segment)
                    edge.set_direction(skel.nodes[ul], skel.nodes[ll])
                    lft.addEdge(edge)

                if j == (m-1):
                    segment = skel.segments[
                        utils.canonicalorder(skel.nodes[lr], skel.nodes[ur])]
                    edge = SkeletonEdge(segment)
                    edge.set_direction(skel.nodes[lr], skel.nodes[ur])
                    rgt.addEdge(edge)

                if i == 0:
                    segment = skel.segments[
                        utils.canonicalorder(skel.nodes[ll], skel.nodes[lr])]
                    edge = SkeletonEdge(segment)
                    edge.set_direction(skel.nodes[ll], skel.nodes[lr])
                    btm.addEdge(edge)

                if i == (n-1):
                    segment = skel.segments[
                        utils.canonicalorder(skel.nodes[ur], skel.nodes[ul])]
                    edge = SkeletonEdge(segment)
                    edge.set_direction(skel.nodes[ur], skel.nodes[ul])
                    top.addEdge(edge)
                el.findHomogeneityAndDominantPixel(skel)
                if pBar.query_stop():
                    pBar.set_failure()
                    pBar.set_message("Failed")
                    return None
                else:
                    rectangle_count = i*m+j
                    pBar.set_progress(float(rectangle_count)/tot_items)
                    pBar.set_message("Creating elements: %d/%d" %\
                                     (rectangle_count, tot_items) )
        if not pBar.query_stop():
            pBar.set_message("Succeeded")
            pBar.set_success()
        return skel
Registration('QuadSkeleton', SkeletonGeometry, QuadSkeleton, 0,
             tip="A Skeleton of quadrilateral elements.")
    

class TriSkeleton(SkeletonGeometry):
    def __init__(self, arrangement=moderate):
        SkeletonGeometry.__init__(self, 'tri')
        self.arrangement = arrangement
    def __call__(self, m, n, microStructure, preset_homog=0):
        pBar = progressbar.getProgress()
        skel = Skeleton(microStructure)
        btm = skel.getEdgeBoundary('bottom', exterior=1)
        top = skel.getEdgeBoundary('top', exterior=1)
        rgt = skel.getEdgeBoundary('right', exterior=1)
        lft = skel.getEdgeBoundary('left', exterior=1)

        toplft = skel.getPointBoundary('topleft', exterior=1)
        toprgt = skel.getPointBoundary('topright', exterior=1)
        btmlft = skel.getPointBoundary('bottomleft', exterior=1)
        btmrgt = skel.getPointBoundary('bottomright', exterior=1)

        ## create nodes
        dx = (skel.MS.size()[0]*1.0)/m
        dy = (skel.MS.size()[1]*1.0)/n
        pBar.set_message("allocating nodes")
        tot_items = (m + 1)*(n + 1)
        for i in range(n+1):
            for j in range(m+1):
                node = skel.newNode(j*dx, i*dy)
                if i==0 and j==0:
                    btmlft.addNode(node)
                if i==0 and j==m:
                    btmrgt.addNode(node)
                if i==n and j==0:
                    toplft.addNode(node)
                if i==n and j==m:
                    toprgt.addNode(node)
                if pBar.query_stop():
                    pBar.set_failure()
                    pBar.set_message("Failed")
                    return None
                else:
                    pBar.set_progress(1.0*(i*(n+1)+j)/tot_items)
        ## create elements and edges
        pBar.set_message("creating elements from nodes")
        tot_items = m*n
        rightdiag = 1
        for i in range(n):                  # loop over rows of elements
            for j in range(m):              # loop over columns of elements
                ll = i*(m+1) + j            # lower left node index
                lr = i*(m+1) + j+1          # lower right
                ur = (i+1)*(m+1) + j+1      # upper right
                ul = (i+1)*(m+1) + j        # upper left
                #
                if self.arrangement == conservative:
                    rightdiag = 1
                elif self.arrangement == liberal:
                    rightdiag = 0
                elif self.arrangement == moderate:
                    rightdiag = (i+j)%2
                elif self.arrangement == middling:
                    rightdiag = 1-(i+j)%2
                elif self.arrangement == anarchic:
                    rightdiag = random.choice([0,1])
                else:
                    debug.fmsg('unknown arrangement!', self.arrangement)


                #  ul  _________  ur
                #      |       /|
                #      | el1  / |
                #      |     /  |
                #      |    /   |
                #      |   /    |
                #      |  /     |
                #      | /  el2 |
                #  ll  |/_______| lr
                #
                if rightdiag==1:
                    el1 = skel.newElement([skel.nodes[ll],skel.nodes[ur],
                                           skel.nodes[ul]])
                    el2 = skel.newElement([skel.nodes[ur],skel.nodes[ll],
                                           skel.nodes[lr]])
                #
                #  ul  _________  ur
                #      |\       |
                #      | \      |
                #      |  \ el2 |
                #      |   \    |
                #      |    \   |
                #      |     \  |
                #      | el1  \ |
                #  ll  |_______\| lr
                #
                else:
                    el1 = skel.newElement([skel.nodes[lr],skel.nodes[ul],
                                           skel.nodes[ll]])
                    el2 = skel.newElement([skel.nodes[ul],skel.nodes[lr],
                                           skel.nodes[ur]])

                # Simple skeleton -- set the homogeneity to "1".
                if preset_homog:
                    el1.cachedHomogeneity.value = 1.0
                    el1.cachedDominantPixel = microStructure.pixelFromPoint(
                        el1.repr_position())
                    el2.cachedHomogeneity.value = 1.0
                    el2.cachedDominantPixel = microStructure.pixelFromPoint(
                        el2.repr_position())
                    
                # Element constructors make the segments, we can use them
                # to make the edges and add them to the boundaries.
                # This code is identical to the block in quadSkeleton, maybe
                # should be encapsulated somewhere.
                if j == 0:                 
                    segment = skel.segments[
                        utils.canonicalorder(skel.nodes[ul], skel.nodes[ll])]
                    edge = SkeletonEdge(segment)
                    edge.set_direction(skel.nodes[ul], skel.nodes[ll])
                    lft.addEdge(edge)

                if j == (m-1):
                    segment = skel.segments[
                        utils.canonicalorder(skel.nodes[lr], skel.nodes[ur])]
                    edge = SkeletonEdge(segment)
                    edge.set_direction(skel.nodes[lr], skel.nodes[ur])
                    rgt.addEdge(edge)

                if i == 0:
                    segment = skel.segments[
                        utils.canonicalorder(skel.nodes[ll], skel.nodes[lr])]
                    edge = SkeletonEdge(segment)
                    edge.set_direction(skel.nodes[ll], skel.nodes[lr])
                    btm.addEdge(edge)

                if i == (n-1):
                    segment = skel.segments[
                        utils.canonicalorder(skel.nodes[ur], skel.nodes[ul])]
                    edge = SkeletonEdge(segment)
                    edge.set_direction(skel.nodes[ur], skel.nodes[ul])
                    top.addEdge(edge)
                el1.findHomogeneityAndDominantPixel(skel)
                el2.findHomogeneityAndDominantPixel(skel)
                if pBar.query_stop():
                    pBar.set_failure()
                    pBar.set_message("Failed")
                    return None
                else:
                    rectangle_count = i*m+j
                    pBar.set_progress(float(rectangle_count)/tot_items)
                    pBar.set_message("Creating elements: %d/%d" % \
                                     (2*rectangle_count, 2*tot_items) )
        if not pBar.query_stop():
            pBar.set_message("Succeeded")
            pBar.set_success()
        return skel

Registration('TriSkeleton', SkeletonGeometry, TriSkeleton, 1,
             params=[enum.EnumParameter('arrangement', Arrangement, moderate,
                                        tip="How to arrange triangles in Skeleton?")],
             tip='A Skeleton of triangular elements.')

######################

# SkeletonBase is provided just so different kinds of skeletons
# (Skeleton, DeputySkeleton) can be checked for in a single
# isinstance() call.

class SkeletonBase:
    def __init__(self):
        self._illegal = 0


    # Utility function, finds all the intersections of passed-in
    # segment (a primitives.Segment object) with the passed-in
    # skeleton element.  Needs the skeleton object in order to extract
    # skeleton segments. Actually returns a dictionary, indexed by
    # points, whose values are lists of the intersecting segments, as
    # a tuple, (intersection-point, next-element)
    def _get_intersections_with_element(self, local_seg, skel_el):
        skel_segs = skel_el.getSegments(self)
        isec_set = {}
        for s in skel_segs:
            nds = s.nodes()
            c1 = nds[0].position()
            c2 = nds[1].position()
            p1 = primitives.Point(c1[0],c1[1])
            p2 = primitives.Point(c2[0],c2[1])
            seg = primitives.Segment(p1,p2)
            isec = local_seg.intersection(seg)
            if isec:
                try:
                    isec_set[isec].append(s)
                except KeyError:
                    isec_set[isec]=[s]
        return isec_set


    # Element traversal function -- given a skeleton element, returns
    # the furthest point (along the "local_seg", which must be a
    # primtives.Segment object) where the local-segment crosses it,
    # and the next element the local segment enters.
    def get_intersection_and_next_element(
        self, local_seg, skel_el):
        
        # First, see if we're already done -- if the current element
        # encloses the trailing point of the local segment, there is
        # no next element.
        if skel_el.interior(local_seg.end()):
            return (None, None)
        
        isec_set = self._get_intersections_with_element(
            local_seg, skel_el)
        
        # Find the trailing intersection.
        m=0
        for (k,v) in isec_set.items():
            mag=(k-local_seg.start())**2
            if mag > m:
                m=mag
                isec_point = k
                isec_segs = v

        # Now we have an intersection point and one or two segments.
        # If there's only one segment, then just find the element.
        if len(isec_segs)==1:
            next_el = isec_segs[0].getOtherElement(skel_el)
            
        # If there are two segments, then we exit through a corner.
        elif len(isec_segs)==2:
            # Find the node common to the two segments.
            nds1 = isec_segs[0].nodes()
            nds2 = isec_segs[1].nodes()
            for n in nds1:
                if n in nds2:
                    common_node = n
                    break
            # Find all the elements connected to this corner which
            # have new intersections.
            corner_elements = common_node.neighborElements()
            new_elements = []
            for e in [x for x in corner_elements if x!=skel_el]:
                new_isecs = self._get_intersections_with_element(
                    local_seg, e)
                try:
                    # Remove the one we already know about.
                    del new_isecs[isec_point]
                except KeyError:
                    pass

                # Also remove any new intersections which precede
                # isec_point along the local segment.  These can sneak
                # in the case where the local segment intersects two
                # segments in the current element, but only one in the
                # previous element -- in that case, the current
                # element will find an exit intersection in the prior
                # element, and incorrectly select it as next, leading
                # to an infinite loop.
                kill_list = []
                mp = (isec_point-local_seg.start())**2
                for k in new_isecs.keys():
                    mk = (k-local_seg.start())**2
                    if mk < mp:
                        kill_list.append(k)
                for k in kill_list:
                    del new_isecs[k]

                # For elements where valid intersections occur,
                # add them to the list of candidates.
                if len(new_isecs)>0:
                    new_elements.append(e)

            # If we found zero elements, then the cross-section must
            # terminate inside one of these.  Figure out which one
            # by testing interiority.
            if len(new_elements)==0:
                for e in [x for x in corner_elements if x!=skel_el]:
                    if e.interior(local_seg.end()):
                        next_el = e
                        break
                    
            # If we found exactly one element, then just return it.
            elif len(new_elements)==1:
                next_el = new_elements[0]
            else:
                # If there's more than one, then the local segment
                # must pass directly between two elements along their
                # shared skeleton segment.  Pick the one on the right,
                # which is the one whose center has a positive cross
                # product with the segment itself, viewed from the
                # start of the segment.
                for e in new_elements:
                    v1 = e.center()-local_seg.start()
                    v2 = local_seg.end()-local_seg.start()
                    if v1.cross(v2)>0:
                        next_el = e
                        break
        else:
            # Impossible!
            raise ooferror.ErrSetupError(
                "Linear path crosses an element more than twice, or fewer than zero times.")

        return (isec_point, next_el)
            

    def enclosingElement(self, point):
        # Find the element containing the given point.
        # Start at the last element found (if any).  Draw a line
        # between the center of the element and the target point.
        # Find which side of the element intersects the line, and move
        # to the neighboring element across that side.  Repeat.  If no
        # side crosses the line, the current element is the one we're
        # looking for.

        # The algorithm relies on starting inside an element, any
        # element.  We use the last element found because it's likely
        # to be near the element that the user is interested in.
        # However, if the last element is illegal, its inside may not
        # be well defined.  Since we use the element's center to find
        # an interior point, and the center of an illegal element may
        # be outside of the element, we don't start at illegal
        # elements.
        
        # _found_element is either None or a weak reference to the
        # last el found
        el = self._found_element and self._found_element()

        # If the (clicked) point is outside of the skeleton,
        # this scheme doesn't work. So, here's a reasonable fix to that.
        if point.x < 0.0:
            point[0] = 0.0
        if point.x > self.MS.size()[0]:
            point[0] = self.MS.size()[0]
        if point.y < 0.0:
            point[1] = 0.0
        if point.y > self.MS.size()[1]:
            point[1] = self.MS.size()[1]

        # If we don't have a good starting point, we look for one.
        if el is None or el.illegal():
            for ell in self.elements:
                if not ell.illegal():
                    el = ell
                    break
            else:               
                raise ooferror.ErrSetupError("All elements are illegal!")
                         
            
        center = el.center()
        straw = primitives.Segment(center, point)

        while(el):
            last_el = el
            (isec, el) = self.get_intersection_and_next_element(
                straw, last_el)

        self._found_element = weakref.ref(last_el)
        return last_el

        

    def nearestSgmt(self, point):
        # Local function to compute the distance between a point pt
        # and segment.  The return value is a tuple whose first entry
        # is the distance squared from the point to the closest point
        # on the segment.  If the projection of the point onto the
        # line of the segment is *not* on the segment, the second
        # member of the tuple is the distance squared along the line
        # from the projected point to the closest endpoint of the
        # segment.  Simply comparing tuples then gives the closest
        # segment to the given point.
        def distance(pt, segment):
            nodes = segment.nodes()
            p0 = nodes[0].position()
            p1 = nodes[1].position()
            a = pt-p0
            b = p1-p0
            seglength2 = b**2
            f = ((a*b)/seglength2)
            if f < 0:
                alpha = -f
                r = pt - p0
            elif f > 1:
                alpha = f-1
                r = pt - p1
            else:
                r = a-f*b
                alpha = 0
            return (r**2, alpha*alpha*seglength2)

        # First, see if the point is inside an element.  If it is, the
        # closest segment must be one of the element's edges.
        element = self.enclosingElement(point)
        if element is not None:
            mindist = None
            nearseg = None
            for i in range(element.nnodes()):
                node0 = element.nodes[i]
                node1 = element.nodes[(i+1)%element.nnodes()]
                seg = self.getSegment(node0, node1)
                d = distance(point, seg)
                if mindist is None or d < mindist:
                    mindist = d
                    nearseg = seg
        else:
            # Not inside any element.  Search all boundary segments.
            mindist = None
            nearseg = None
            for seg in self.segments.values():
                if seg.nElements() == 1: # it's a boundary segment
                    d = distance(point, seg)
                    if mindist is None or d < mindist:
                        mindist = d
                        nearseg = seg
        return nearseg

    ## Call setIllegal() after creating an illegal element.
    def setIllegal(self):
        self._illegal = 1
    def illegal(self):
        return self._illegal

    ## Call checkIllegality after any operation that may have changed
    ## an illegal skeleton into a legal one.  Operations that change
    ## legal skeletons into illegal ones should be able to check
    ## legality themselves (without searching all elements) and should
    ## call setIllegal() directly.
    def checkIllegality(self):
        for el in self.elements:
            if el.illegal():
                self._illegal = 1
                return
        self._illegal = 0

#####################

# Skeleton objects live in a SkeletonContext stack, and many of their
# operations are invoked via the Context, which does important
# bookkeeppiinngg.

class Skeleton(SkeletonBase):
    """
    The geometrical information for a mesh, without any of the
    complications of nodes, shapefunctions, or materials.
    """
    def __init__(self,microStructure):
        SkeletonBase.__init__(self)
        self.MS = microStructure
        self._size = self.MS.size()
        self._area = self._size[0]*self._size[1]
        self.nodemovehistory = skeletondiff.NodeMoveHistory()
        self.elements = []
        self._found_element = None      # used in enclosingElement().
        self.nodes = []
        self.segments = {}              # Nondirected edges.
        self.edgeboundaries = {}
        self.pointboundaries = {}
        self.timestamp = timestamp.TimeStamp()
        self.meshes = []                # list of Mesh objects for real meshes
        self.node_dictionary = {} ## for fast access. Only used in parallel
        self.element_dictionary ={} ## also for fast access. Only used in parallel
        self.segment_dictionary = {} ## and yet another object for fast access, only used in parallel
        # When elements and nodes are deleted from the mesh, they
        # aren't immediately removed from the lists in the Skeleton.
        # They're only removed when cleanUp() is called.  washMe
        # indicates whether or not cleanUp() is necessary.
        self.washMe = 0
        
        self._needHash = 1

        self.deputy = None              # currently active DeputySkeleton
        self.deputylist = []            # all deputies
        self._deferreddestruction = 0
        self._destroyed = 0

        self.sbcallbacks = [
            switchboard.requestCallback('materials changed in microstructure',
                                        self.matChanged)
            ]
        self._name = ""

    def isDeputy(self):
        return 0

    def name(self):
        return self._name

    def set_name(self,nm):
        self._name = nm
        
    def destroy(self, skelcontext):
        ## NOTE: destroy() may be called more than once, if the
        ## Skeleton has deputies.  If the Skeleton has deputies when
        ## it is destroyed, the _deferreddestruction flag is set, and
        ## destroy() will be called again when the last deputy is
        ## destroyed.  That means that destroy() can't leave lists of
        ## destroyed objects lying around -- it must actually replace
        ## the lists with empty lists.

        map(switchboard.removeCallback, self.sbcallbacks)
        self.sbcallbacks = []           # see NOTE above
        
        # Loop over a copy of the mesh list instead of the list
        # itself, because Mesh.destroy() calls Skeleton.remove_mesh(),
        # which removes items from the list and messes up the
        # iteration.

        meshes = self.meshes[:]
        for mesh in meshes:
            mesh.destroy()              # removes from self.meshes
        self.meshes = []                # see NOTE above
        self._destroyed = 1             # see NOTE above
        
        # Any data shared with deputies must not be deleted until the
        # deputies are done with it.
        if self.ndeputies() == 0:
            for ebdy in skelcontext.edgeboundaries.values():
                ebdy.remove(self)
            for pbdy in skelcontext.pointboundaries.values():
                pbdy.remove(self)
            del self.MS
            self.disconnect()
        else:
            self._deferreddestruction = 1
            
    def destroyed(self):
        return self._destroyed

    def __repr__(self):
        return 'Skeleton(%d)' % id(self)
    
    def disconnect(self):
        for s in self.nodes + self.segments.values() + self.elements:
            s.disconnect()

    def getTimeStamp(self):
        return self.timestamp
##         return max(self.timestamp, self.MS.getTimeStamp())

    def matChanged(self, ms):           # switchboard "materials changed ..."
        if ms is self.MS:
            for mesh in self.meshes:
                mesh.refreshMaterials()

    def hashNodes(self):
        # Computing preliminary information
        nnodes = self.nnodes()  # no. of nodes
        x_size = self.size()[0]
        y_size = self.size()[1]
        ratio = x_size/y_size  # aspect ratio of skeleton
        if ratio < 1.0:
            ratio = 1.0/ratio
        nndtile = int( 0.5*math.sqrt(nnodes)*ratio )  # no. of nodes per tile
        ntiles = int( nnodes/nndtile )  # no. of tiles

        if x_size >= y_size:
            nx = int( math.sqrt(ratio*ntiles) )  # tiles in the x-direction
            ny = int( (1.0*nx)/ratio )           # tiles in the y-direction
        else:
            ny = int( math.sqrt(ratio*ntiles) )
            nx = int( (1.0*ny)/ratio )

        # For large ratios, the int can yield zero.  Fix these.
        if nx==0: nx=1
        if ny==0: ny=1
        
        self.hashedNodes = skeletonnode.HashedNodes((nx,ny), self.size())

        self.hashedNodes.hash(self)
        self._needHash = 0


    def needsHash(self):
        self._needHash = 1

    def nnodes(self):
        self.cleanUp()
        return len(self.nodes)
                                   
    def nelements(self):                # for compatiblity w/ Element output
        self.cleanUp()
        return len(self.elements)

    def element_iterator(self):         # for compatiblity w/ Element output
        self.cleanUp()
        return self.elements

    def node_iterator(self):
        self.cleanUp()
        return self.nodes

    def segment_iterator(self):
        self.cleanUp()
        return self.segments.values()

    def getNodeIndex(self, node):
        return self.nodes.index(node)

    def notPinnedNodes(self):
        return [n for n in self.node_iterator() if not n.pinned()]
##        nodes = []
##        for n in self.node_iterator():
##            if not n.pinned():
##                nodes.append(n)
##        return nodes

    def getElementIndex(self, elem):
        return self.elements.index(elem)

    def nillegal(self):
        n = 0
        for e in self.elements:
            if e.illegal():
                n += 1
        return n

    # Returns a tuple containing maximum x-extent and maximum y-extent
    # of the skeleton. 
    def size(self):
        return self._size

    def area(self):
        return self._area
        
    def newNode(self, x, y):
        c = SkeletonNode(x,y)
        if x == 0.0 or x == self.size()[0]:
            c.setMobilityX(0)
        if y == 0.0 or y == self.size()[1]:
            c.setMobilityY(0)

        self.nodes.append(c)
        return c

    # Only elements with nonzero area are constructed here.  The
    # line-elements are expected to be done in the boundary code.
    def newElement(self, nodes, parents=[]):
        nnodes = len(nodes)
        if nnodes==3:
            el = SkeletonTriangle(nodes)
        elif nnodes==4:
            el = SkeletonQuad(nodes)
        else:
            raise RuntimeError(
                "Unable to construct %d-noded element." % nnodes)
        self.elements.append(el)
        for parent in parents:
            el.add_parent(parent)
            parent.add_child(el)
      
        # Add this element's edges to the dictionary of segments.
        # getSegment returns an existing SkeletonSegment, or makes one
        # if necessary.
        lastnode = nodes[-1]
        for node in nodes:
            segment = self.getSegment(lastnode, node)
            segment.addElement(el)
            lastnode = node

        return el

##    def loadNode(self, x, y):
##        return self.newNode(x, y)

    def loadElement(self, *indices):
        nodes = [self.nodes[index] for index in indices]
        return self.newElement(nodes)

    def loadEdge(self, node0, node1):
        seg = self.getSegment(node0, node1)
        edge = skeletonsegment.SkeletonEdge(seg)
        if seg.nodes()[0] == node0:
            edge.direction = 1
        else:
            edge.direction = -1
        return edge

    def removeElements(self, *elements):
        self.washMe = 1
        for el in elements:
            el.defunct = 1
            el.destroy(self)

    def removeNode(self, node):
        # Called only by SkeletonNode.destroy() which is called by
        # SkeletonNode.removeElement() when the node's last element is
        # removed.
        self.washMe = 1
        node.defunct = 1

    def cleanUp(self):
        if self.washMe:
            utils.weedlist(lambda e: not hasattr(e, 'defunct'), self.elements)
            utils.weedlist(lambda n: not hasattr(n, 'defunct'), self.nodes)
            self.washMe = 0

    def getElement(self, index):
        return self.elements[index]

    def getNode(self, index):
        return self.nodes[index]
        
    # getSegment returns an existing segment joining the given nodes,
    # or creates a segment if such a segment doesn't already exist.
    def getSegment(self, node0, node1):
        nodes = utils.canonicalorder(node0, node1)
        try:
            return self.segments[nodes]
        except KeyError:
            segment = SkeletonSegment(node0, node1)
            self.segments[nodes] = segment
            return segment

    # findSegment returns an existing segment joining the given nodes,
    # or None if such as segment doesn't exist.
    def findSegment(self, node0, node1):
        try:
            return self.segments[utils.canonicalorder(node0, node1)]
        except KeyError:
            return None

    def removeSegment(self, key):
        # Called only by SkeletonSegment.destroy(), which is called by
        # SkeletonSegment.removeElement when the segment's last
        # element is removed.
        del self.segments[key]

    def clone(self,name):
        return self.properCopy()

    def properCopy(self):
        """
        Copies the current skeleton properly so that the new skeleton and the
        current skeleton are totally independent.
        """
        
        self.cleanUp()
        # create a new skeleton
        newSkeleton=Skeleton(self.MS)

        # Make new nodes which have the same indices, but are children
        # of the old nodes.
        newSkeleton.nodes = [ n.copy_child() for n in self.nodes ] 
        newSkeleton.elements = [ e.copy_child() for e in self.elements ]

        # Make new segments which have equivalent indices (with new
        # nodes), but are children of the old segments.
        for s in self.segments.values():
            new_seg = s.copy_child()
            newSkeleton.segments[new_seg.nodes()] = new_seg

        newSkeleton._illegal = self._illegal
        return newSkeleton
            

    def improperCopy(self):
        # Copy a Skeleton, but *not* the elements or segments.  Just
        # nodes.  Used when refining, where the elements and segments
        # are recreated by Refine.apply().
        self.cleanUp()
        newSkeleton = Skeleton(self.MS)
        newSkeleton.nodes = [n.copy_child() for n in self.nodes]
        newSkeleton._illegal = self._illegal
        return newSkeleton

    ###################

    # The following routines are redefined in the DeputySkeleton
    # class.  A DeputySkeleton is a skeleton that differs from another
    # skeleton only in the position of its nodes.  See
    # engine/deputy.py. 

    def deputyCopy(self):
        from engine import deputy       # delayed import to avoid loops
        # A "copy" that doesn't actually make a copy, but just keeps
        # track of which nodes have been moved.
        return deputy.DeputySkeleton(self)

    def sheriffSkeleton(self):          # The sheriff isn't the deputy
        return self

    def deputize(self, deputy):         # install a new deputy
        if self.deputy:
            self.deputy.deactivate()
        self.deputy = deputy

    def addDeputy(self, dep):
        # Called by DeputySkeleton.__init__()
        self.deputylist.append(dep)

    def removeDeputy(self, dep, skelcontext):
        # Called by DeputySkeleton.destroy()
        self.deputylist.remove(dep)
        if self._deferreddestruction and self.ndeputies() == 0:
            self.destroy(skelcontext)

    def ndeputies(self):
        return len(self.deputylist)

    def activate(self):
        if self.deputy:
            self.deputy.deactivate()
            self.deputy = None

    def moveNodeTo(self, node, position):
        node.moveTo(position)

    def moveNodeBy(self, node, delta):
        node.moveBy(delta)

    def moveNodeBack(self, node):
        node.moveBack()

    def getMovedNodes(self):
        return {}

    def nodePosition(self, node):
        # Gets the position of the node in this skeleton even if a
        # deputy is active.
        if self.deputy:
            return self.deputy.originalPosition(node)
        return node.position()
            
    def newSelectionTracker(self, selectionset):
        return skeletonselectable.SelectionTracker()

    def newGroupTracker(self, groupset):
        return skeletongroups.GroupTracker()

    def newPinnedNodeTracker(self):
        return skeletonnode.PinnedNodeTracker(self)

    def promoteTrackers(self, context):
        pass

    #######################
    def weightedEnergyTotal(self, alpha):
        e = 0.0
        self.cleanUp()
        for element in self.elements:
            e += element.area()*element.energyTotal(self, alpha)
        return e
    def energyTotal(self, alpha):
        e = 0.0
        self.cleanUp()
        for element in self.elements:
            e += element.energyTotal(self, alpha)
        return e

    def illegalElements(self):
        return [e for e in self.elements if e.illegal()]
##        # return a COPY of illegal element list.
##        # TODO: why?  The original list isn't being used for anything
##        illegal = []
##        for element in self.elements:
##            if element.illegal():
##                illegal.append(element)
##        return illegal[:]

    def activeElements(self):
        self.cleanUp()
        return [e for e in self.elements if e.active(self)]
##        activeelements = []
##        for element in self.elements:
##            if element.active(self):
##                activeelements.append(element)
##        return activeelements

    def activeNodes(self):
        self.cleanUp()
        return [n for n in self.nodes if n.active(self)]
##        activenodes = []
##        for node in self.nodes:
##            if node.active(self):
##                activenodes.append(node)
##        return activenodes

    def activeSegments(self):
        self.cleanUp()
        return [s for s in self.segments.values() if s.active(self)]
##        activesegments = []
##        for segment in self.segments.values():
##            if segment.active(self):
##                activesegments.append(segment)
##        return activesegments
                    
    def nearestNode(self, point):
        if self._needHash:
            self.hashNodes()
        return self.hashedNodes.nearestNode(point)


    def mergeNode(self, node0, node1):
        # Create a provisional skeleton change object representing the
        # merging of nodes node0 and node1 at the position of node1.
        if not node0.canMergeWith(node1):
            return
        doomedSegment = self.findSegment(node0, node1)
        if doomedSegment is None:
            return
        topElements = doomedSegment.getElements()
        isoElements = [elephant for elephant in node0.neighborElements()
                       if elephant not in topElements]
        
        node0.moveTo(node1.position())
        for element in isoElements:
            if element.illegal():
                node0.moveBack()
                return
        # Make sure to move the node back to where it was, because the
        # merge may still be rejected.
        node0.moveBack()
            
        change = ProvisionalMerge(node0,node1)
        for oldelement in isoElements:
            newnodes = oldelement.nodes[:]
            newnodes[newnodes.index(node0)] = node1
            change.substituteElement(
                oldelement,
                skeletonelement.getProvisionalElement(newnodes,
                                                      oldelement.getParents()))
        for oldelement in topElements:
            change.removeElements(oldelement)
            if oldelement.nnodes() > 3:
                newnodes = oldelement.nodes[:]
                newnodes.remove(node0)
                newelement = skeletonelement.getProvisionalElement(
                    newnodes, parents=oldelement.getParents())
                change.insertElements(newelement)
        return change

    ########################################################################

    def getPointBoundary(self, name, exterior=None):
        try:
            return self.pointboundaries[name]
        except KeyError:
            if exterior:
                bdy = ExteriorSkeletonPointBoundary(name)
            else:
                bdy = SkeletonPointBoundary(name)
            self.pointboundaries[name] = bdy
            return bdy

    def getEdgeBoundary(self, name, exterior=None):
        try:
            return self.edgeboundaries[name] # existing bdy with this name
        except KeyError:                 # didn't find existing bdy
            if exterior:
                bdy = ExteriorSkeletonEdgeBoundary(name)
            else:
                bdy = SkeletonEdgeBoundary(name)       # create it
            self.edgeboundaries[name] = bdy  # save it
            return bdy

    # The SkeletonContext has already ensured that there is no collision.
    # This routine is called from the SkelContextBoundary object's
    # rename routine.
    def renameBoundary(self, oldname, newname):
        if oldname in self.edgeboundaries.keys():
            self.edgeboundaries[newname]=self.edgeboundaries[oldname]
            del self.edgeboundaries[oldname]
        elif oldname in self.pointboundaries.keys():
            self.pointboundaries[newname]=self.pointboundaries[oldname]
            del self.pointboundaries[oldname]

        # Context-level function, because boundary conditions need
        # to remain consistent.
        for mctxt in self.meshes:
            mctxt.renameBoundary(oldname, newname)

    # Build a new edge boundary from the passed-in list of segments,
    # and return it.  The boundary should "point" from the first
    # segment to the last.  Startnode is required if there is only
    # one segment, and is ignored in the other cases.
    #
    # Caller must provide a topologically trivial list of segments
    # with length greater than zero, so all we have to do here is
    # figure out the directions for the edges.
    def makeEdgeBoundary(self, name, segments=None, startnode=None,
                         exterior=None):
        if (name in self.edgeboundaries.keys()) or \
               (name in self.pointboundaries.keys()):
            raise ooferror.ErrPyProgrammingError(
                "Boundary '%s' already exists." % name)
        
        bdy = self.getEdgeBoundary(name, exterior) # Guaranteed to be new.
        
        if segments==None: 
            return bdy # Return "empty" boundary -- used for stack propagation.

        if len(segments)==1:
            if startnode:
                seg = segments[0]
                if startnode==seg.nodes()[0]:
                    bdy.addEdge(SkeletonEdge(seg, 1))
                else: # startnode==seg.nodes()[1]:
                    bdy.addEdge(SkeletonEdge(seg, -1))
            else:
                raise ooferror.ErrPyProgrammingError(
                    "Singleton segment boundaries require a starting node!")

        else: # Length of the segment list is greater than one.
            for i in range(len(segments)-1):
                seg1 = segments[i]
                seg2 = segments[i+1]
                if seg1.nodes()[0] in seg2.nodes():
                    bdy.addEdge(SkeletonEdge(seg1, -1))
                else: #  seg1.nodes()[1] in seg2.nodes():
                    bdy.addEdge(SkeletonEdge(seg1, 1))
            # For the final segment, need to check the one previous.
            seg1 = segments[-2]
            seg2 = segments[-1]
            if seg2.nodes()[0] in seg1.nodes():
                bdy.addEdge(SkeletonEdge(seg2, 1))
            else: #  seg2.nodes()[1] in seg1.nodes():
                bdy.addEdge(SkeletonEdge(seg2, -1))

        # Write this boundary to any meshes we have.
        for mctxt in self.meshes:
            mctxt.newEdgeBoundary(name, bdy)
                
        return bdy
            

    # Build a new point boundary from the passed-in list of nodes,
    # and return it.  
    def makePointBoundary(self, name, nodes=None, exterior=None):
        if (name in self.pointboundaries.keys()) or \
               (name in self.edgeboundaries.keys()):
            raise ooferror.ErrPyProgrammingError(
                "Boundary '%s' already exists." % name)

        bdy = self.getPointBoundary(name, exterior)

        # Correctly returns an empty boundary if nodes==None.
        if nodes:
            for n in nodes:
                bdy.addNode(n)

        for mctxt in self.meshes:
            mctxt.newPointBoundary(name, bdy)

        return bdy
    

    def removeBoundary(self, name):
        if name in self.pointboundaries.keys():
            del self.pointboundaries[name]
        elif name in self.edgeboundaries.keys():
            del self.edgeboundaries[name]

        # Remove it from the meshes.
        for mctxt in self.meshes:
            # Let the mesh *context* do it, because that's where the
            # boundary conditions live.
            mctxt.removeBoundary(name)


    # The named boundary has been modified -- change the versions in
    # the mesh to match.  Don't just remove and replace, as this
    # destroys valuable boundary condition info.
    def pushBoundaryToMesh(self, name):
        for mctxt in self.meshes:
            if name in self.pointboundaries.keys():
                b = self.pointboundaries[name]
                mctxt.replacePointBoundary(name, b)
            elif name in self.edgeboundaries.keys():
                b = self.edgeboundaries[name]
                mctxt.replaceEdgeBoundary(name, b)

    def mapBoundary(self, bdy, skeleton, **kwargs):
        # double dispatch wrapper for SkelContextBoundary.map().

        # Copy boundary information from the given skeleton to this
        # skeleton.  The given skeleton might be a deputy, which
        # doesn't have any boundary information, so actually copy from
        # the deputy's sheriff.

        # mapBoundary is a no-op in the DeputySkeleton class, so we
        # don't have to worry about copying boundary data *to* a
        # deputy.  However, if the given source skeleton is a deputy,
        # there's a chance that this skeleton is its sheriff, in which
        # case we don't actually have to copy anything.
        omar = skeleton.sheriffSkeleton()
        if omar is not self.sheriffSkeleton():
            bdy.map(omar, self, **kwargs)

##    def rmBoundary(self, bdy):
##        # double dispatch wrapper for SkelContextBoundary.remove()
##        bdy.remove(self)

    def find_geometrical_boundaries(self):
        # reset
        nbrdict = {}
        for el in self.elements:
            el.exterior_edges = []

        # Count neighbors of each skeleton element, and keep a dictionary
        # listing the elements that share each edge.
        for el in self.elements:
            for i in range(el.nnodes()):   # loop over edges
                key = utils.canonicalorder(el.nodes[i],
                                           el.nodes[(i+1)%el.nnodes()])
                try:                    # see if this edge has already been seen
                    nbrdict[key].append(el)
                except KeyError:        # The edge hasn't been seen yet
                    nbrdict[key] = [el]

        # Look for edges that only belong to one element.
        for edge,elist in nbrdict.items(): # loop over edges
            if len(elist)==1:           # there's only one
                el = elist[0]
                el.exterior_edges.append(edge)

    ##############################

    def quick_sanity_check(self):
        # Just check for illegal elements.  For a more thorough check,
        # see sanity_check(), below.
        sane = 1
        area = 0.
        for element in self.elements:
            area += element.area()
            if element.illegal():
                reporter.report("illegal element", element.index,
                                [n.position() for n in element.nodes])
                sane = 0
        reporter.report("Total element area is", area)
        reporter.report("Microstructure area is", self.MS.area())
        if sane:
            reporter.report("*** The toe bone's connected to the foot bone. ***")
        else:
            reporter.report("*** Skeleton sanity check failed. ***")
        return sane

    def sanity_check(self):
        sane = 1
        for element in self.elements:
            if element.illegal():
                reporter.report("illegal element", element.index,
                                [n.position() for n in element.nodes])
                sane = 0
            for node in element.nodes:
                if node not in self.nodes:
                    reporter.report("element", element.index, "contains a node",
                                    node.index, "not in the skeleton")
                    sane = 0
            segs = element.getSegments(self)
            if None in segs:
                reporter.report("Element", element.index,
                                "is missing a segment")
                sane = 0
        for node in self.nodes:
            for element in node.neighborElements():
                if element not in self.elements:
                    reporter.report("node", node.index, "contains an element",
                                    element.index, "not in the skeleton")
                    sane = 0
        for segment in self.segments.values():
            elements = segment.getElements()
            if len(elements) > 2:
                reporter.report("segment", [n.index for n in segment.nodes()], 
                                "has too many elements:",
                                [el.index for el in elements])
                sane = 0
            for element in elements:
                if element not in self.elements:
                    reporter.report("segment",
                                    [n.index for n in segment.nodes()],
                                    "contains an element", element.index, 
                                    "not in the skeleton")
                    sane = 0
            for node in segment.nodes():
                if node not in self.nodes:
                    reporter.report("segment",
                                    [n.index for n in segment.nodes()], 
                                    "contains a node", node.index,
                                    "not in the skeleton")
                    sane = 0
            
        if sane:
            reporter.report(
                "*** The ankle bone's connected to the shin bone. ***")
        else:
            reporter.report("*** Skeleton sanity check failed. ***")
        return sane

                
# ## ### #### ##### ###### ####### ######## ####### ###### ##### #### ### ## #


    # The "mesh" function creates a real mesh with pre-assigned edge
    # boundaries 'top', 'bottom', 'left', and 'right', and preassigned
    # single point boundaries 'topleft', 'topright', 'bottomleft' and
    # 'bottomright'.  Anything else, you gotta make for yourself.

    def femesh(self, edict, set_materials):
        "Create a real mesh from a Skeleton, using the given element types."
        # edict[n] is the n-sided master element.
        ## debug.fmsg(edict)
        self.cleanUp()

        # Local dictionary of finite-element nodes, indexed by
        # SkeletonNode objects.
        fe_node = {}
        
        # Find which elements and edges are on the geometrical
        # boundaries of the system.
        self.find_geometrical_boundaries()

        realmesh = femesh.FEMesh()

        # make the real nodes at the nodes of the elements.  These
        # nodes are always both mapping and function nodes.  A
        # "meshindex" attribute is written into the skeleton node.
        for i in range(len(self.nodes)):
            cur = self.nodes[i]
            realnode = realmesh.newFuncNode(coord.Coord(cur.position().x,
                                                        cur.position().y))
            fe_node[cur] = realnode
            cur.setMeshIndex(realnode.index())

            
        mnodecount = len(self.nodes)
        edgedict = {}                   # Edge dictionary
        # Loop over elements.
        for mesh_idx in range(len(self.elements)):
            el = self.elements[mesh_idx]

            # Index correspondence happens here -- the skeleton
            # elements are assigned indices in the order that their
            # corresponding real elements are created/assigned. 
            mnodecount += el.realelement(self, realmesh, mesh_idx, fe_node,
                                         mnodecount, edgedict, edict,
                                         set_materials)
            
        # Then do boundaries.
        # Note that edgeboundaries and pointboundaries are in separate lists
        # in the skeleton, but in a single list in the real mesh.

        # Point boundaries first.
        for bdkey, pointbndy in self.pointboundaries.items():
            realbndy = realmesh.newPointBoundary(bdkey)
            for node in pointbndy.nodes:
                realbndy.addNode(fe_node[node]) # Preserve order of nodes.

        # ... then edge boundaries.
        for bdkey,edgebndy in self.edgeboundaries.items():
            edgebndy.sequence()
            realbndy = realmesh.newEdgeBoundary(bdkey)
            for skeletonedge in edgebndy.edges:
                # print "Doing an edge."
                
                # Look up the corresponding element from the skeleton.
                skelel = skeletonedge.segment.getElements()[0]
                realel = realmesh.getElement(skelel.meshindex)
                
                edge_nodes = skeletonedge.get_nodes()
                realn0 = fe_node[edge_nodes[0]]    # First real node.
                realn1 = fe_node[edge_nodes[1]]    # Second real node.
                realbndy.addEdge(realel.getBndyEdge(realn0,realn1))
                
        return realmesh

    # Add a Mesh object to our list.  This Mesh object is the wrapper
    # for the FEMesh object that was created by Skeleton.femesh(),
    # above.
    def add_mesh(self, mesh):
        self.meshes.append(mesh)

    def remove_mesh(self, mesh):
        self.meshes.remove(mesh)

    def mark_for_parallelization(self):
        ## this function partitions the mesh into
        ## sub-regions, in parallel mode. Each region
        ## sets its ownership
        pass

##    def set_real_materials(self, mnodecount, realmesh, fe_node, edgedict, edict):
##        for mesh_idx in range(len(self.elements)):
##            el = self.elements[mesh_idx]
            
##            # Index correspondence happens here -- the skeleton
##            # elements are assigned indices in the order that their
##            # corresponding real elements are created/assigned. 
##            mnodecount += el.realelement(self, realmesh, mesh_idx, fe_node,
##                                         mnodecount, edgedict, edict,  el.material)
    
    def set_skeleton_materials(self, mnodecount, realmesh, fe_node, edgedict, edict):
        pass

    
## end of class Skeleton

########################################################################


def newEmptySkeleton(name, msname):
    ms = microstructure.microStructures[msname].getObject()
    skel = Skeleton(ms)
    skeletoncontext.skeletonContexts.add([msname, name], skel)
    return skel

# skeleton_geometry is an object of type SkeletonGeometry, class defined above.
def initialSkeleton(name, nx, ny, microStructure, skeleton_geometry):
    skel = skeleton_geometry(nx, ny, microStructure) ## progressbar control occurs here
    if skel is not None:

        skeletoncontext.skeletonContexts.add([microStructure.name(), name], skel)
    return skel

# Creats pixel-to-element skeleton. Thus, homogeneity of all elements will
# be set to "1".
def simpleSkeleton(name, nx, ny, microStructure, skeleton_geometry):
    skel = skeleton_geometry(nx, ny, microStructure, preset_homog=1)
    skeletoncontext.skeletonContexts.add([microStructure.name(), name], skel)
    return skel

# Export skeleton-creation functions to the restricted environment.
from common import utils
utils.OOFdefine('Skeleton', Skeleton)
utils.OOFdefine('initialSkeleton',initialSkeleton)

###########################

class ProvisionalChanges:
    def __init__(self):
        self.removed = []               # elements to be removed 
        self.inserted = []              # provisional elements added
        self.substitutions = []         # pairs (old, new) of el. substitutions
        self.seg_subs = {}              # {old:[new segs] ...}
        self.movednodes = []            # list of MoveNode objects
        self.cachedDeltaE = None        # Energy difference
        self.before = None              # Elements before the change
        self.after = None               # Elements after the chnage
            
    class MoveNode:                     # nested class definition
        def __init__(self, node=None, position=None, mobility=None):
            self.node = node
            self.position = position
            self.mobility = mobility

    def removeElements(self, *elements):
        for element in elements:
            self.removed.append(element)

    def insertElements(self, *elements):
        for element in elements:
            self.inserted.append(element)

    def substituteElement(self, old, new):
        # Old and new elements must have the same number of nodes, and
        # corresponding nodes must be in the same positions in the
        # element's node lists, so that the correct parent-child
        # relationships may be made for the corresponding segments.
        self.substitutions.append([old, new])

    def substituteSegment(self, old, new):  # "new" has to be a list
        self.seg_subs[old] = new

    def moveNode(self, node, position, mobility=(1,1)):
        self.movednodes.append(
            self.MoveNode(node=node, position=position, mobility=mobility))

    def nRemoved(self):
        return len(self.removed)

    def elBefore(self):
        if self.before is None:
            self.before = self.removed + [o for o,n in self.substitutions]
            for mvnode in self.movednodes:
                for nbr in mvnode.node.neighborElements():
                    if nbr not in self.before:
                        self.before.append(nbr)
        return self.before

    def elAfter(self):
        if self.after is None:
            self.after = self.inserted + [n for o,n in self.substitutions]
            for mvnode in self.movednodes:
                for nbr in mvnode.node.neighborElements():
                    if (nbr not in self.removed) and (nbr not in self.after):
                        self.after.append(nbr)
        return self.after

    def makeNodeMove(self, skeleton):
        for mvnode in self.movednodes:
            mvnode.node.moveTo(mvnode.position)

    def moveNodeBack(self, skeleton):
        for mvnode in self.movednodes:
            mvnode.node.moveBack()        

    def illegal(self, skeleton):
        # Will this change produce any illegal elements?
        verboten = 0
        # Move nodes accordingly to simulate the change
        self.makeNodeMove(skeleton)
        # Check elements
        for element in self.elAfter():
            if element.illegal():
                verboten = 1
                break
        # Move nodes back
        self.moveNodeBack(skeleton)
        return verboten

    def deltaE(self, skeleton, alpha):
        # Return the change in energy per element if this move were to
        # be accepted.
        if self.cachedDeltaE is None:
            # Energy before the change
            oldE = 0.0
            for element in self.elBefore():
                oldE += element.energyTotal(skeleton, alpha)
            oldE /= len(self.elBefore())
            # Move nodes accordingly to simulate the change
            self.makeNodeMove(skeleton)
            # Energy after the change
            newE = 0.0
            for element in self.elAfter():
                newE += element.energyTotal(skeleton, alpha)
            newE /= len(self.elAfter())
            # Move node back
            self.moveNodeBack(skeleton)
            # Energy differnce due to the change
            self.cachedDeltaE = newE - oldE
        return self.cachedDeltaE

##     def illegal(self, skeleton):
##         # Will this change produce any illegal elements?
##         verboten = 0
##         elements = self.inserted + [new for old,new in self.substitutions]
##         for mvnode in self.movednodes:
##             mvnode.node.moveTo(mvnode.position)
##             for nbr in mvnode.node.neighborElements():
##                 if (nbr not in self.removed) and (nbr not in elements):
##                     elements.append(nbr)

##         for element in elements:
##             if element.illegal():
##                 verboten = 1
##                 break

##         for mvnode in self.movednodes:
##             mvnode.node.moveBack()

##         return verboten

##     def deltaE(self, skeleton, alpha):
##         # Return the change in energy per element if this move were to
##         # be accepted.
##         if self.cachedDeltaE is not None:
##             return self.cachedDeltaE
        
##         elBefore = self.removed + [old for old,new in self.substitutions]
##         elAfter = self.inserted + [new for old,new in self.substitutions]
##         # Find elements affected by node motion, being careful not to
##         # double count.  The neighbor elements of a moving node may be
##         # removed, or may be neighbors of more than one moving node.
##         for mvnode in self.movednodes:
##             for element in mvnode.node.neighborElements():
##                 # lists aren't likely to be long, so naive search is ok.
##                 if element not in elBefore:
##                     elBefore.append(element)
##                     elAfter.append(element)
                    
##         oldE = 0.0
##         if elBefore:
##             for element in elBefore:
##                 oldE += element.energyTotal(skeleton, alpha)
##             oldE /= len(elBefore)
            
##         for mvnode in self.movednodes:
##             mvnode.node.moveTo(mvnode.position)

##         newE = 0.0
##         if elAfter:
##             for element in elAfter:
##                 newE += element.energyTotal(skeleton, alpha)
##             newE /= len(elAfter)

##         for mvnode in self.movednodes:
##             mvnode.node.moveBack()

##         self.cachedDeltaE = newE - oldE
##         return self.cachedDeltaE

    def accept(self, skeleton):
        # Create actual elements to replace the provisional ones.  The
        # actual elements replace their predecessors in the
        # ProvisionalChanges object, so that they're available to the
        # calling routine.
        self.inserted = [element.accept(skeleton) for element in self.inserted]
        for mvnode in self.movednodes:
            mvnode.node.moveTo(mvnode.position)
            if mvnode.mobility:
                mvnode.node.setMobilityX(mvnode.mobility[0])
                mvnode.node.setMobilityY(mvnode.mobility[1])
        for pair in self.substitutions:
            old, new = pair
            newelement = new.accept(skeleton)
            pair[1] = newelement
            oldsegments = old.getSegments(skeleton)
            newsegments = newelement.getSegments(skeleton)
            for oldseg, newseg in zip(oldsegments, newsegments):
                for parent in oldseg.getParents():
                    newseg.add_parent(parent)
                    parent.add_child(newseg)
            # Call Skeleton.removeElements only *after* the segment
            # parents have been reestablished, because removing the
            # elements may remove the segments from the skeleton.
            skeleton.removeElements(old)
        for old in self.seg_subs.keys():
            new_segs = self.seg_subs[old]
            for new in new_segs:
                for parent in old.getParents():
                    new.add_parent(parent)
                    parent.add_child(new)
        skeleton.removeElements(*self.removed)

class ProvisionalInsertion(ProvisionalChanges):
    def __init__(self):
        ProvisionalChanges.__init__(self)
        self.addedNodes = []

    def addNode(self, node):
        self.addedNodes.append(node)
        
    def removeAddedNodes(self, skeleton):
        for n in self.addedNodes:
            n.destroy(skeleton)

class ProvisionalMerge(ProvisionalChanges):
    def __init__(self, node0, node1):
        ProvisionalChanges.__init__(self)
        self.node0 = node0
        self.node1 = node1
        
    def accept(self, skeleton):
        self.node0.makeSibling(self.node1)
        ProvisionalChanges.accept(self, skeleton)

## class ProvisionalDoubleMerge(ProvisionalChanges):
##     def __init__(self, n0, n1, n2, n3):
##         ProvisionalChanges.__init__(self)
##         self.n0 = n0
##         self.n1 = n1
##         self.n2 = n2
##         self.n3 = n3

##     def accept(self, skeleton):
##         self.n0.makeSibling(self.n1)
##         self.n2.makeSibling(self.n3)
##         ProvisionalChanges.accept(self, skeleton)
