# -*- python -*-
# $RCSfile: skeletoncontext.py,v $
# $Revision: 1.103 $
# $Author: langer $
# $Date: 2005/02/25 19:56:56 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# A SkeletonContext is a subclass WhoDoUndo class.


from SWIG.common import switchboard
from SWIG.common import timestamp
from SWIG.common import ooferror
from common.IO import whoville
from common import debug
from common import parallel_enable
from common import utils
from engine import skeletonselectable
from engine import skeletongroups
from engine import skeletonboundary
from engine import skeletonnode
from engine.IO import movenode
from common import microstructure
import weakref
import string

# When propagating boundaries, Deputies shouldn't be around - since they
# don't have nothing to do with boundaries!
# Takes a list of skeletons and removes Deputies.
def withoutDeputies(skels):
    return [skel for skel in skels if not skel.isDeputy()]
##    omars = []
##    for skel in skels:
##        if not skel.isDeputy():
##            omars.append(skel)
##    return omars

class SkeletonContext(whoville.WhoDoUndo):
    def __init__(self, name, classname, skel):
        # All of the args have to be given, although not all are used.
        # This is because the constructor is called by WhoClass.add,
        # which assumes that the arguments are the same as those of
        # Who.__init__.  Probably a better solution using
        # RegisteredClasses could be found.
        self._skelname = name

        self.edgeboundaries = {}
        self.pointboundaries = {}
        self.selectedBdyName = None
        self.bdyselected = timestamp.TimeStamp()

        # Various selection objects live here.  Instance the
        # selections from their constructors.  To get the current
        # selection, do "context.elementselection.retrieve()".
        self.nodeselection = skeletonselectable.NodeSelection(self)
        self.segmentselection = skeletonselectable.SegmentSelection(self)
        self.elementselection = skeletonselectable.ElementSelection(self)
        self.pinnednodes = skeletonnode.PinnedNodeSelection(self)

        # These attribute names (nodegroups, segmentgroups,
        # elementgroups) are used in the generic menu callback in
        # IO/skeletongroupmenu.py.  Also in skeletonselectionmod.py.
        # Change them in all places (or none, of course.)
        self.nodegroups = skeletongroups.NodeGroupSet(self)
        self.segmentgroups = skeletongroups.SegmentGroupSet(self)
        self.elementgroups = skeletongroups.ElementGroupSet(self)
        
        # Init calls pushModification, so timing is important.
        # pushModification calls implied_select for all the
        # selections, so the selection objects have to exist at this
        # point.
        whoville.WhoDoUndo.__init__(self, name, 'Skeleton', skel)

        self.requestCallback(('preremove who', 'Microstructure'),
                             self.removeWho)

        # Overwrite function, gets called when an item in the stack is
        # about to be overwritten.  This assignment replaces  the default
        # overwrite function set by the parent class initializer.
        self.undobuffer.overwrite = self.skeletonDestroy

        # Ensure that the passed-in obj really does start a new "family."
        skel.disconnect()

        # Ask the initial skel about any pre-existing boundaries.
        for (name, bdy) in skel.edgeboundaries.items():
            self.edgeboundaries[name] = \
              bdy.makeContextBoundary(self, name, skel)

        for (name, bdy) in skel.pointboundaries.items():
            self.pointboundaries[name] = \
              bdy.makeContextBoundary(self, name, skel)

    # Remove must be sure to break circular references.
    def remove(self):
        self.nodeselection.destroy()
        self.segmentselection.destroy()
        self.elementselection.destroy()
        self.nodegroups.destroy()
        self.segmentgroups.destroy()
        self.elementgroups.destroy()
        self.pinnednodes.destroy()
        del self.nodeselection
        del self.segmentselection
        del self.elementselection
        del self.pinnednodes
        del self.nodegroups
        del self.elementgroups
        del self.segmentgroups
        # WhoDoUndo.remove destroys all the objects in the undobuffer.
        whoville.WhoDoUndo.remove(self)

    def pushModification(self, skeleton):
        old = self.getObject()
        skeleton.activate()
        # Call the parent pushModification without emitting the "who
        # changed" switchboard signal.
        whoville.WhoDoUndo.pushModification(self, skeleton, signal=0)
        ## parallel push is implicitly applied in the line above
        self.undobuffer[0].promoteTrackers(self)

        # Propagate the boundaries.  This must be done before the
        # boundaries are drawn, which is why the "who changed" signal
        # was suppressed.
        for e in self.edgeboundaries.values():
            skeleton.mapBoundary(e, old) # Default direction is child-ward.
        for p in self.pointboundaries.values():
            skeleton.mapBoundary(p, old) # Default direction is child-ward.

        # Now emit the "who changed' signal.
        self.pushModificationSignal()
        
        switchboard.notify("new boundary configuration", self)
        skeleton.set_name(skeleton.MS.name() + ':' + self._skelname)

    # Extra operations during WhoDoUndo undo and redo.  If either the
    # old object or the new one is a DeputySkeleton, the positions of
    # the nodes must be updated.  Also, the number of edges/nodes in
    # the current boundary can be changed.
    
    
    def undoHookFn(self, oldskel, newskel):
        newskel.activate()
        self.pinnednodes.replace(oldskel, newskel)
        switchboard.notify("new boundary configuration", self)

    def redoHookFn(self, oldskel, newskel):
        newskel.activate()
        self.pinnednodes.replace(oldskel, newskel)
        switchboard.notify("new boundary configuration", self)

    def promoteDeputyTrackers(self, deputyskeleton):
        # Called by DeputySkeleton.promoteTrackers, but not by
        # Skeleton.promoteTrackers, which are called by
        # SkeletonContext.pushModification.
        self.elementselection.promoteDeputyTrackers(deputyskeleton)
        self.nodeselection.promoteDeputyTrackers(deputyskeleton)
        self.segmentselection.promoteDeputyTrackers(deputyskeleton)
        self.nodegroups.promoteDeputyTracker(deputyskeleton)
        self.elementgroups.promoteDeputyTracker(deputyskeleton)
        self.segmentgroups.promoteDeputyTracker(deputyskeleton)
##        self.pinnednodes.promoteDeputyTrackers(deputyskeleton)
        
    # ## ### #### ##### ###### ####### ####### ###### ##### #### ### ## #

    # Boundary functions -- create, destroy, eventually edit.
    # To create a boundary, create it in the current skeleton, then
    # propagate it upward and downward.  Likewise to destroy.
    def createEdgeBoundary(self, name, segment_set, startnode,
                           exterior=None, autoselect=1):
        bdy = self.getObject().makeEdgeBoundary(name,
                                                segment_set, startnode,
                                                exterior)
        context_bdy = bdy.makeContextBoundary(self, name, self.getObject())
        self.edgeboundaries[name] = context_bdy
        
        # Propagation up and down.
        # Unfortunate nomenclature -- in the context, mapping "down"
        # is towards the children, which is the direction of
        # increasing index in the undobuffer, and is therefore towards
        # the top.
        current = self.getObject()
        children = self.undobuffer.getToTop()
        for c in withoutDeputies(children[1:]):
            c.mapBoundary(context_bdy, current,
                          direction=skeletonboundary.MAP_DOWN)
            current = c

        current = self.getObject()
        parents = self.undobuffer.getToBottom()
        for p in withoutDeputies(parents[1:]):
            p.mapBoundary(context_bdy, current,
                          direction=skeletonboundary.MAP_UP)
            current = p

        switchboard.notify("new boundary created", self)
        if autoselect:
            self.selectBoundary(name)

    def createPointBoundary(self, name, node_set,
                            exterior=None, autoselect=1):
        bdy = self.getObject().makePointBoundary(name, node_set, exterior)
        context_bdy = bdy.makeContextBoundary(self, name, self.getObject())
        self.pointboundaries[name] = context_bdy
        
        # Propagation up and down.
        # See routine above for nomenclature lament.
        # DeputySkeleton.mapBoundary is a no-op.
        current = self.getObject()
        children = self.undobuffer.getToTop()
        for c in withoutDeputies(children[1:]):
            c.mapBoundary(context_bdy, current,
                          direction=skeletonboundary.MAP_DOWN)
            current = c

        current = self.getObject()
        parents = self.undobuffer.getToBottom()
        for p in withoutDeputies(parents[1:]):
            p.mapBoundary(context_bdy, current,
                          direction=skeletonboundary.MAP_UP)
            current = p

        switchboard.notify("new boundary created", self)
        if autoselect:
            self.selectBoundary(name)

    # The edgeboundaries[name] or pointboundaries[name] objects here
    # are SkelContextBoundary objects -- their "remove" methods
    # remove the corresponding Skeleton[Point|Edge]Boundary objects
    # from the skeletons.
    def removeBoundary(self, name):
        # Actual removal from skeletons happens in the
        # SkelContext<type>Boundary's remove routine.
        if name in self.edgeboundaries.keys():
            self.edgeboundaries[name].remove()
            del self.edgeboundaries[name]
        elif name in self.pointboundaries.keys():
            self.pointboundaries[name].remove()
            del self.pointboundaries[name]
        else:
            raise ooferror.ErrUserError(
                "Cannot remove boundary %s, no such boundary." % name)
        if name == self.selectedBdyName:
            self.unselectBoundary()
        switchboard.notify("boundary removed", self)

    def renameBoundary(self, oldname, newname):
        if oldname==newname:
            return
        
        if newname in self.allBoundaryNames():
            raise ooferror.ErrSetupError("Name %s already in use." % newname)
         
        if oldname in self.edgeboundaries.keys():
            dict = self.edgeboundaries
        elif oldname in self.pointboundaries.keys():
            dict = self.pointboundaries
        else: 
            raise ooferror.ErrPyProgrammingError(
                "Boundary name %s not found." % oldname)

        dict[newname]=dict[oldname]
        del dict[oldname]
        dict[newname].rename(newname)
        switchboard.notify("boundary renamed", self)
        self.selectBoundary(newname)

    # Retrieval routine -- returns the SkeletonContextBoundary object.
    def getBoundary(self, name):
        try:
            return self.edgeboundaries[name]
        except KeyError:
            return self.pointboundaries[name]

    # The selected boundary isn't actually used for anything directly,
    # but it's displayed in the GUI and its name is passed in as an
    # argument to commands that operate on boundaries.  
    def selectBoundary(self, bdyname):
        self.selectedBdyName = bdyname
        self.bdyselected.increment()    # timestamp
        switchboard.notify("boundary selected", self, bdyname)
        switchboard.notify('redraw')
    def unselectBoundary(self):
        if self.selectedBdyName is not None:
            self.selectedBdyName = None
            self.bdyselected.increment()
            switchboard.notify("boundary unselected", self)
            switchboard.notify('redraw')
    def getSelectedBoundaryName(self):
        return self.selectedBdyName
    def getSelectedBoundary(self):      # returns SkeletonContextBoundary object
        if self.selectedBdyName is not None:
            return self.getBoundary(self.selectedBdyName)
            

    # ## ### #### ##### ###### ####### ####### ###### ##### #### ### ## #

    # Routines that modify a boundary in place.  All of them exit via
    # the "bdyPropagate", which propagates the modified boundaries up
    # and down the stack.  All of these routines call
    # SkelContextBoundary objects which actually operate on the
    # sheriff of the current skeleton, not the current skeleton
    # itself.


    def appendSegstoBdy(self, name, seg_set):
        try:
            bdy = self.edgeboundaries[name]
        except KeyError:
            raise ooferror.ErrSetupError(
                "Boundary %s is not an edge boundary.", name)

        bdy.append(seg_set)

        self.bdyPropagate(name, bdy)

    def reverseBoundary(self, name):
        try:
            bdy = self.edgeboundaries[name] # SkelContextEdgeBoundary
        except KeyError:
            raise ooferror.ErrSetupError("Boundary %s is not an edge boundary.",
                                         name)
        bdy.reverse()
        self.bdyPropagate(name, bdy)


    def addNodestoBdy(self, name, node_list):
        try:
            bdy = self.pointboundaries[name]
        except KeyError:
            raise ooferror.ErrSetupError(
                "Boundary %s is not a point boundary.", name)

        bdy.append(node_list)

        self.bdyPropagate(name, bdy)


    # Segment list has been sequenced by the time we get here, and are
    # from the current .getObject().
    def removeSegsfromBdy(self, name, seg_set):
        try:
            bdy = self.edgeboundaries[name]
        except KeyError:
            raise ooferror.ErrSetupError(
                "Boundary %s is not an edge boundary.", name)

        bdy.delete(seg_set)
        
        self.bdyPropagate(name, bdy)

        
    def removeNodesfromBdy(self, name, node_list):
        try:
            bdy = self.pointboundaries[name]
        except KeyError:
            raise ooferror.ErrSetupError(
                "Boundary %s is not a point boundary.", name)

        bdy.delete(node_list)

        self.bdyPropagate(name, bdy)

    def bdyPropagate(self, name, contextbdy):
        # Replace the boundaries up and down the stack with a suitably
        # mapped version of the modified boundary.  Operate only
        # on sheriffs, and don't re-operate on the original object.

        # Then, ensure that the modifications get to the meshes --
        # in this case, we need to run over the whole stack, because
        # although deputy skeletons don't have boundaries, their
        # meshes nevertheless do.

        start = self.getObject().sheriffSkeleton() # Current sheriff
        children = self.undobuffer.getToTop()

        current = start
        for c in withoutDeputies(children[1:]):
            if not c is start:
                contextbdy.remove(c)
                contextbdy.map(current, c,
                               direction=skeletonboundary.MAP_DOWN)
                current = c

        current = start
        parents = self.undobuffer.getToBottom()
        for p in withoutDeputies(parents[1:]):
            if not p is start:
                contextbdy.remove(p)
                contextbdy.map(current, p, direction=skeletonboundary.MAP_UP)
                current = p

        # Now loop over the whole undo buffer, breathe
        # regularly, and push.
        for skel in self.undobuffer:
            skel.pushBoundaryToMesh(name)

        switchboard.notify("new boundary configuration", self)


    # ## ### #### ##### ###### ####### ####### ###### ##### #### ### ## #

    def allEdgeBoundaryNames(self):
        return self.edgeboundaries.keys()
    
    def allBoundaryNames(self):
        return self.edgeboundaries.keys() + self.pointboundaries.keys()

    def uniqueBoundaryName(self, name):
        return utils.uniqueName(name, self.allBoundaryNames())

    # Get information about the named boundary, i.e. type, size, and
    # return it as a string, with newlines as appropriate.
    def boundaryInfo(self, boundaryname):
        outlist = []
        type = "Edge"
        bdy = None
        try:
            bdy = self.edgeboundaries[boundaryname]
        except KeyError:
            type = "Point"
            try:
                bdy = self.pointboundaries[boundaryname]
            except KeyError:
                type = "Unknown"

        outlist.append("Type: %s" % type)

        if bdy:
            outlist.append("Size: %i" % bdy.size(self.getObject()))
        
        return string.join(outlist,"\n")
        
                
                

    # ## ### #### ##### ###### ####### ####### ###### ##### #### ### ## #

    # This function is passed a reference to a skeleton that is about
    # to be overwritten in the stack.  The boundaries need to be
    # removed from the skeleton to break circular references between
    # the edges and segments, and all the selectables should unhook
    # themselves from their parents and/or children.
    def skeletonDestroy(self, skeleton):
        skeleton.destroy(self)


    # Copy over the group data from another skeletoncontext.
    def groupCopy(self, other):
        mygroups = [self.nodegroups, self.segmentgroups, self.elementgroups]
        othergroups = [other.nodegroups, other.segmentgroups,
                       other.elementgroups] 
        for (g, og) in map(None, mygroups, othergroups):
            g.nameCopy(og)

    # ## ### #### ##### ###### ####### ####### ###### ##### #### ### ## #

    # List retrievers, from here to the end of the undo buffer,
    # including (in both cases) the current skeleton.

    def getChildList(self):
        return self.undobuffer.getToTop()

    def getParentList(self):
        return self.undobuffer.getToBottom()

    # ## ### #### ##### ###### ####### ####### ###### ##### #### ### ## #    
    
    def mesh(self):
        return self.getObject()

    def getMicrostructure(self):
        return self.getObject().MS
    
    def removeWho(self, whopath):
        if whopath[-1] == self.getObject().MS.name():
            # Our Microstructure has been removed.  Remove ourself.
            skeletonContexts.remove(self.path())

    def getSelectionContext(self, mode):
        return mode.getSelectionContext(self)
            
    def __repr__(self):
        return 'SkeletonContext("%s")' % self.name()


    def parallelize(self):
        global skeletonContexts
        if parallel_enable.query() == 1:
            try:
                from engine.IO import skeletonIPC
                from SWIG.common import mpitools
                if mpitools.rank() == 0:
                    skeletonIPC.smenu.Initialize(skeleton =skeletonContexts.getPath(self))
            except ImportError:
                pass
        return

    def parallelPushModification(self, obj=None):
        ## assume top-most image is going to be pushed to the back-end
        global skeletonContexts
        if parallel_enable.query() == 1:
            try:
                from common import parallel_object_manager
                from engine.IO import skeletonIPC
                from SWIG.common import mpitools
                if mpitools.rank() == 0:
                    try:
                        ## send name and ID to look-up the REAL skeleton in the back-end
                        _id = parallel_object_manager.parallelObjectManager.getIndex(obj)
                        skeletonIPC.smenu.Push(skeleton =skeletonContexts.getPath(self), id = _id)
                        ## remove skeleton from object Manager in front-end
                        ## back-end makes it share of work on the back
                        parallel_object_manager.parallelObjectManager.delete(_id)
                    except KeyError:
                        pass
            except ImportError:
                pass
        return
    
    def parallelUndoModification(self):
        global skeletonContexts
        if parallel_enable.query() == 1:
            try:
                from engine.IO import skeletonIPC
                from SWIG.common import mpitools
                if mpitools.rank() == 0:
                    skeletonIPC.smenu.Undo(skeleton =skeletonContexts.getPath(self))
            except ImportError:
                pass
        return
    def parallelRedoModification(self):
        global skeletonContexts
        if parallel_enable.query() == 1:
            try:
                from engine.IO import skeletonIPC
                from SWIG.common import mpitools
                if mpitools.rank() == 0:
                    skeletonIPC.smenu.Redo(skeleton =skeletonContexts.getPath(self))
            except ImportError:
                pass
        return


    
skeletonContexts = whoville.WhoDoUndoClass(
    'Skeleton',
    ordering=200,
    parentClass=microstructure.microStructures,
    instanceClass=SkeletonContext,
    proxyClasses=['<topmost>'])

utils.OOFdefine('skeletonContexts', skeletonContexts)

import engine.mesh

def latestMesh():                       # a vestigial function?
    latestmesh = None
    latesttime = timestamp.TimeStamp()
    latesttime.backdate()
    for context in (skeletonContexts, engine.mesh.meshes):
        for path in context.keys():
            mesh = context[path]
            ts = mesh.getTimeStamp()
            if ts > latesttime:
                latesttime = ts
                latestmesh = mesh
    return latestmesh

#############################################

def getSkeleton(ms, name):
    # ms is a Microstructure.  Returns a SkeletonContext
    return skeletonContexts[[ms.name(), name]]
