// -*- C++ -*-
// $RCSfile: solverstepper.C,v $
// $Revision: 1.24 $
// $Author: redwing $
// $Date: 2005/03/01 16:31:13 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modifed
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>
#include "engine/csrmatrix.h"
#include "engine/femesh.h"
#include "engine/matvec.h"
#include "engine/solver.h"
#include "engine/solverstepper.h"

double dot_product(VECTOR_D &left, VECTOR_D &right)
{
  double result = 0;
  for (int i = 0 ; i<left.size(); i++)
    result += left(i)*right(i);
  return result;

}

SolverStepperCore::~SolverStepperCore() {
  //std::cerr << "destroying SolverStepperCore object" << std::endl;
  clear_memory();
  //std::cerr << "leaving SolverStepperCore destroy call" << std::endl;
}

SolverStepperCore::SolverStepperCore(LinearSolver * solver, FEMesh * femesh) 
  : solver(solver),
    femesh(femesh),
    A(0),
    C(0),
    x(0),
    rhs(0),
    xp(0),
    nrowsA(0), 
    ncolsA(0)
{

}

double SolverStepperCore::residual()
{
  //same residual as the one used by the SparseLib solvers
  return solver->get_residual();
  
}

void SolverStepperCore::precompute() {
  //std::cerr << "SolverStepperCore::precomputing"  << std::endl;
  //at this point the solver and the preconditioner are already selected
  // at the UI and passed as arguments into the solver
  RWLock * rwlock = femesh->get_rwlock();
  rwlock->read_acquire(); //acquire read lock
  //std::cerr << "after the lock" << std::endl;
  try{
    int n, m;
    if (A)
      {
	delete A;
      }
    A = new CSRmatrix (femesh->Amatrix()); 
    
    nrowsA = n = A->nrows();
    
    // std::cerr << "SolverStepperCore::precompute: " << A->nrows() << "x" << A->ncols() << std::endl;
    ncolsA = A->ncols();
    if(n != A->ncols())
      throw ErrSetupError("Stiffness matrix is not square! nrows="
			  + to_string(n) + ", ncols="
			  + to_string(A->ncols()));
      
    if (C)
      {
	delete C;
      }
    C = new CSRmatrix (femesh->Cmatrix());
    m = C->ncols();
    if (!x)
    {
      x = new VECTOR_D (n, 0.0);
    }
    x->clear();
    femesh->get_unknowns(*x, femesh->freedofmap);
    //std::cerr << "x=" << *x << std::endl;
    if (!rhs)
    {
      ////std::cerr << "allocating memory for rhs" << std::endl;
      rhs = new VECTOR_D (n, 0.0);
    }
    rhs->clear();
    femesh->get_rhs(*rhs, femesh->indepeqnmap);
    
    if (!xp)
    {
    //std::cerr << "allocating memory for xp" << std::endl;
      xp = new VECTOR_D (m, 0.0);
    }
    xp->clear();
    //std::cerr << "allocating memory for xp" << std::endl;
    femesh->get_unknowns(*xp, femesh->fixeddofmap);
    *rhs -= (*C) * (*xp);

    
  } //end of try block
  catch(...) //guarantee that lock will be released if ANY exception is raised
    {
      //std::cerr << "releasing the lock in catch" << std::endl;
      rwlock->read_release(); 
      throw;
    }
  //std::cerr << "releasing the lock while exiting the function " << std::endl;
  rwlock->read_release();

  
  
}


bool SolverStepperCore::solve(){
  //  std::cerr << "entering olverStepperCore::solve" << std::endl;
  solver->set_blocking(femesh);
  solver->precompute(*A);
  //std::cerr << "A = "<< *A << std::endl;
  bool status = solver->solve(*A, *rhs, *x);
  solver->postcompute(); // Call possibly-null post-compute hook.
  postcompute();
  return status;
  
}

void SolverStepperCore::postcompute()
{
  //std::cerr << "SolverStepperCore::postcomputing"  << std::endl;
  RWLock * rwlock = femesh->get_rwlock();
  rwlock->write_acquire();
 try
   {
     femesh->set_unknowns(*x, femesh->freedofmap);
     CSRmatrix Cp = femesh->Cpmatrix();
     CSRmatrix B = femesh->Bmatrix();
     femesh->set_rhs(Cp * (*x) + B * (*xp), femesh->depeqnmap); // <-- not modified to reflect local_rhs.
   }
 catch(...)
   {
     rwlock->write_release();
   }
 rwlock->write_release();
  
}

void SolverStepperCore::clear_memory()
{
  delete  A;
  delete C;
  delete x;
  delete rhs;
  delete xp;

}

NewtonSolverCore::NewtonSolverCore(LinearSolver * solver, FEMesh * femesh)
  : SolverStepperCore(solver, femesh),
    psi(0),
    dx(0)
{

}

NewtonSolverCore::~NewtonSolverCore()
{
  //std::cerr << "destroying NR SolverCore" << std::endl;
  delete psi; //free memory of psi
  delete dx;
  //std::cerr << "leaving NR destructor" << std::endl;
}

double NewtonSolverCore::residual()
{
  //std::cerr << "inside NewtonSolverCore::residual()" << std::endl;
  //same residual as the one used by the SparseLib solvers
  if (!psi || !rhs)
    {
      
      return 0;
    }
  double psi_mag = dot_product(*psi, *psi);
  double rhs_mag = dot_product(*rhs, *rhs);
  //std::cerr << "psi=" << psi_mag << " rhs=" << rhs_mag << std::endl;
  if (rhs_mag == 0.0) 
    rhs_mag = 1;
  //std::cerr << "exiting NewtonSolverCore::residual()" << std::endl;
  return sqrt(psi_mag/rhs_mag);
}

void NewtonSolverCore::precompute()
{
  //std::cerr << "entering NR::precompute" << std::endl;
  SolverStepperCore::precompute();
  RWLock * rwlock = femesh->get_rwlock();
  rwlock->read_acquire(); //acquire read lock
  try{
    if (!psi)
      {
	//std::cerr << "allocating memory for psi" << std::endl;
	//at this point the fields have been
	//initialized, so it is okay to
	//build psi
	psi = new VECTOR_D (A->nrows(), 0.0);
      }
    *psi = (*rhs) - (*A) * (*x); 
    VECTOR_D temp_v = (*A) * (*x);
    //std::cerr << "x=" << dot_product(*x,*x) << std::endl;
    //std::cerr << "A*x=" << dot_product(temp_v,temp_v) << std::endl;
    //std::cerr << "rhs=" << dot_product(*rhs,*rhs) << std::endl;
    //std::cerr << "psi=" << dot_product(*psi,*psi) << std::endl;
    //std::cerr << "A = "<< *A << std::endl;
    if (dx)
      delete dx;
    //
    //if (!dx)
    // {
    //std::cerr << "allocating memory for dx" << std::endl;
    dx = new VECTOR_D (A->nrows(), 0.0);
    
    //
    //}
    
  }
  catch(...)
    {
      rwlock->read_release(); 
      throw;
    }
  rwlock->read_release(); 
}


bool NewtonSolverCore::solve()
{
  //std::cerr << "inside NewtonSolverCore::solve()" << std::endl;
  solver->set_blocking(femesh);
  //std::cerr << "after set_blocking" << std::endl;
  solver->precompute(*A);
  //std::cerr << "after precomputing A" << std::endl;
  bool status = solver->solve(*A, *psi, *dx);
  //std::cerr << "after solver->solve()" << std::endl;
  if (status)
    (*x) += (*dx); 
  solver->postcompute(); // Call post-compute hook.
  //std::cerr << "after solver postcomputing" << std::endl;
  postcompute();
  //std::cerr << "after postcompute" << std::endl;
  return status;
}

void NewtonSolverCore::postcompute()
{
  SolverStepperCore::postcompute();
}
